import math

import numpy as np


def cdog(im_input, im_mask, sigma_min, sigma_max, num_octave_levels=3):
    """SCale-adaptive Multiscale Difference-of-Gaussian (DoG) filter for
    nuclei/blob detection.

    Computes the maximal DoG response over a series of scales where in the
    applicable scales at each pixel are constrained to be below an upper-bound
    equal to 2 times the distance to the nearest non-nuclear/background pixel.

    This function uses an approach similar to SIFT interest detection
    where in the scale space between the specified min and max sigma values is
    divided into octaves (scale/sigma is doubled after each octave) and each
    octave is divided into sub-levels. The gaussian images are downsampled by 2
    at the end of each octave to keep the size of convolutional filters small.

    Parameters
    ----------
    im_input : array_like
        A hematoxylin intensity image obtained from ColorDeconvolution. Objects
        are assumed to be dark with a light background.
    mask : array_like
        A binary image where nuclei pixels have non-zero values
    sigma_min : double
        Minimum sigma value for the scale space. For blob detection, set this
        equal to minimum-blob-radius / sqrt(2).
    sigma_max : double
        Maximum sigma value for the scale space. For blob detection, set this
        equal to maximum-blob-radius / sqrt(2).
    num_octave_levels : int
        Number of levels per octave in the scale space.

    Returns
    -------
    im_dog_max : array_like
        An intensity image containing the maximal DoG response across
        all scales for each pixel
    im_sigma_max : array_like
        An intensity image containing the sigma value corresponding to the
        maximal LoG response at each pixel. The nuclei/blob radius for
        a given sigma value can be calculated as sigma * sqrt(2).


    References
    ----------
    .. [*] D.G. Lowe "Distinctive image features from scale-invariant
           keypoints." International journal of computer vision, vol. 60,
           no. 2, 91-110, 2004.

    """
    from scipy.ndimage import distance_transform_edt, gaussian_filter
    from skimage.transform import resize

    im_input = im_input.astype(float)

    # generate distance map
    im_dmap = distance_transform_edt(im_mask)

    # compute max sigma at each pixel as 2 times the distance to background
    im_sigma_ubound = 2.0 * im_dmap

    # clip max sigma values to specified range
    im_sigma_ubound = np.clip(im_sigma_ubound, sigma_min, sigma_max)

    # compute number of levels in the scale space
    sigma_ratio = 2 ** (1.0 / num_octave_levels)

    k = int(math.log(float(sigma_max) / sigma_min, sigma_ratio)) + 1

    # Compute maximal DoG filter response across the scale space
    sigma_cur = sigma_min
    im_gauss_cur = gaussian_filter(im_input, sigma_cur)
    im_sigma_ubound_cur = im_sigma_ubound.copy()

    MIN_FLOAT = np.finfo(im_input.dtype).min

    im_dog_max = np.zeros_like(im_input)
    im_dog_max[:, :] = MIN_FLOAT
    im_dog_octave_max = im_dog_max.copy()

    im_sigma_max = np.zeros_like(im_input)
    im_sigma_octave_max = np.zeros_like(im_input)

    n_level = 0
    n_octave = 0

    for i in range(k + 1):

        # calculate sigma at next level
        sigma_next = sigma_cur * sigma_ratio

        # Do cascaded convolution to keep convolutional kernel small
        # G(a) * G(b) = G(sqrt(a^2 + b^2))
        sigma_conv = np.sqrt(sigma_next ** 2 - sigma_cur ** 2)
        sigma_conv /= 2.0 ** n_octave

        im_gauss_next = gaussian_filter(im_gauss_cur, sigma_conv)

        # compute DoG
        im_dog_cur = im_gauss_next - im_gauss_cur

        # constrain response
        im_dog_cur[im_sigma_ubound_cur < sigma_cur] = MIN_FLOAT

        # update maxima
        max_update_pixels = np.where(im_dog_cur > im_dog_octave_max)

        if len(max_update_pixels[0]) > 0:

            im_dog_octave_max[max_update_pixels] = im_dog_cur[max_update_pixels]
            im_sigma_octave_max[max_update_pixels] = sigma_cur

            # print np.unique(im_sigma_octave_max)

        # update cur sigma
        sigma_cur = sigma_next
        im_gauss_cur = im_gauss_next

        # update level
        n_level += 1

        # Do additional processing at the end of each octave
        if i == k or n_level == num_octave_levels:

            # update maxima
            if num_octave_levels > 0:

                im_dog_octave_max_rszd = resize(
                    im_dog_octave_max, im_dog_max.shape, order=0)

            else:

                im_dog_octave_max_rszd = im_dog_octave_max

            max_pixels = np.where(
                im_dog_octave_max_rszd > im_dog_max)

            if len(max_pixels[0]) > 0:

                im_dog_max[max_pixels] = \
                    im_dog_octave_max_rszd[max_pixels]

                if num_octave_levels > 0:

                    im_sigma_octave_max_rszd = resize(
                        im_sigma_octave_max, im_dog_max.shape, order=0)

                else:

                    im_sigma_octave_max_rszd = im_sigma_octave_max

                im_sigma_max[max_pixels] = \
                    im_sigma_octave_max_rszd[max_pixels]

            # downsample images by 2 at the end of each octave
            if n_level == num_octave_levels:

                im_gauss_cur = im_gauss_next[::2, ::2]
                im_sigma_ubound_cur = im_sigma_ubound_cur[::2, ::2]

                im_dog_octave_max = im_dog_octave_max[::2, ::2]
                im_sigma_octave_max = im_sigma_octave_max[::2, ::2]

                n_level = 0
                n_octave += 1

    # set min vals to min response
    im_dog_max[im_dog_max == MIN_FLOAT] = 0

    return im_dog_max, im_sigma_max



import os
import shutil
import sys
import tempfile

import pandas as pd
import pytest
import sqlalchemy as db
from sqlalchemy import text

from histomicstk.annotations_and_masks.annotation_database_parser import (
    dump_annotations_locally, parse_annotations_to_local_tables)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
import htk_test_utilities as utilities  # noqa
from htk_test_utilities import getTestFilePath, girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.folderid = None


cfg = Cfg()

# pytest runs tests in the order they appear in the module


@pytest.mark.usefixtures('girderClient')  # noqa
def test_prep(girderClient):  # noqa

    cfg.gc = girderClient

    # get original item
    iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

    # create the folder to "back up"
    folderinfo = cfg.gc.post(
        '/folder', data={
            'parentId': iteminfo['folderId'],
            'name': 'test-parser',
        })
    cfg.folderid = folderinfo['_id']

    # create subfolder to test recursion
    subf = cfg.gc.post(
        '/folder', data={
            'parentId': cfg.folderid,
            'name': 'test-parser-sub',
        })

    # copy the item multiple times to create dummy database
    for i in range(2):
        for fid in (cfg.folderid, subf['_id']):
            _ = cfg.gc.post(
                '/item/%s/copy' % iteminfo['_id'], data={
                    'name': 'test_dbsqlite-%d' % i,
                    'copyAnnotations': True,
                    'folderId': fid,
                })


class TestDatabaseParser:
    """Test girder database parser."""

    def test_dump_annotations_locally_1(self):
        """Test dump annotations locally."""
        savepath = tempfile.mkdtemp()

        # recursively save annotations -- JSONs + sqlite for folders/items
        dump_annotations_locally(
            cfg.gc, folderid=cfg.folderid, local=savepath,
            save_json=True, save_sqlite=True)

        assert not len({
            'test-parser.json',
            'test-parser.sqlite',
            'test_dbsqlite-0.json',
            'test_dbsqlite-0_annotations.json',
            'test_dbsqlite-1.json',
            'test_dbsqlite-1_annotations.json',
            'test-parser-sub',
        } - set(os.listdir(savepath)))

        sql_engine = db.create_engine('sqlite:///%s/test-parser.sqlite' % savepath)
        dbcon = sql_engine.connect()

        result = pd.read_sql_query(text("""SELECT count(*) FROM 'folders';"""), dbcon)
        assert int(result.loc[0, :]) == 2

        result = pd.read_sql_query(text("""SELECT count(*) FROM 'items';"""), dbcon)
        assert int(result.loc[0, :]) == 4

        # cleanup
        shutil.rmtree(savepath)

    def test_dump_annotations_locally_2(self):
        """Test dump annotations locally."""
        savepath = tempfile.mkdtemp()

        # recursively save annotations -- parse to csv + sqlite
        dump_annotations_locally(
            cfg.gc, folderid=cfg.folderid, local=savepath,
            save_json=False, save_sqlite=True,
            callback=parse_annotations_to_local_tables,
            callback_kwargs={
                'save_csv': True,
                'save_sqlite': True,
            },
        )

        assert not len({
            'test-parser.sqlite',
            'test_dbsqlite-0_docs.csv',
            'test_dbsqlite-0_elements.csv',
            'test_dbsqlite-1_docs.csv',
            'test_dbsqlite-1_elements.csv',
            'test-parser-sub'} - set(os.listdir(savepath)))

        files = list(os.listdir(os.path.join(savepath, 'test-parser-sub')))
        assert len([j for j in files if j.endswith('_docs.csv')]) == 2
        assert len([j for j in files if j.endswith('_elements.csv')]) == 2

        sql_engine = db.create_engine('sqlite:///%s/test-parser.sqlite' % savepath)
        dbcon = sql_engine.connect()

        result = pd.read_sql_query(text(
            """SELECT * FROM 'annotation_docs';"""), dbcon)
        assert result.shape == (32, 13)
        assert set(result.columns) == {
            '_modelType', '_version', 'annotation_girder_id',
            'created', 'creatorId', 'element_count',
            'element_details', 'groups', 'itemId', 'item_name',
            'public', 'updated', 'updatedId'}

        result = pd.read_sql_query(text(
            """SELECT * FROM 'annotation_elements';"""), dbcon)
        assert result.shape == (304, 13)
        assert set(result.columns) == {
            'annotation_girder_id', 'bbox_area', 'color',
            'coords_x', 'coords_y', 'element_girder_id', 'group', 'label',
            'type', 'xmax', 'xmin', 'ymax', 'ymin'}

        # cleanup
        shutil.rmtree(savepath)



import large_image
import numpy as np
import PIL.Image

from .simple_mask import simple_mask


def sample_pixels(slide_path, sample_fraction=None, magnification=None,
                  tissue_seg_mag=1.25, min_coverage=0.1, background=False,
                  sample_approximate_total=None, tile_grouping=256, invert_image=False,
                  style=None, frame=None, default_img_inversion=False):
    """Generates a sampling of pixels from a whole-slide image.

    Useful for generating statistics or Reinhard color-normalization or
    adaptive deconvolution. Uses mixture modeling approach to focus
    sampling in tissue regions.

    Parameters
    ----------
    slide_path : str
        path and filename of slide.
    sample_fraction : double
        Fraction of pixels to sample. Must be in the range [0, 1].
    magnification : double
        Desired magnification for sampling.
        Default value : None (for native scan magnification).
    tissue_seg_mag: double, optional
        low resolution magnification at which foreground will be segmented.
        Default value = 1.25.
    min_coverage: double, optional
        minimum fraction of tile covered by tissue for it to be included
        in sampling. Ranges between [0,1). Default value = 0.1.
    background: bool, optional
        sample the background instead of the foreground if True. min_coverage
        then refers to the amount of background. Default value = False
    sample_approximate_total: int, optional
        use instead of sample_fraction to specify roughly how many pixels to
        sample. The fewer tiles are excluded, the more accurate this will be.
    tile_grouping: int, optional
        Number of tiles to process as part of a single task.

    Returns
    -------
    pixels : array_like
        A Nx3 matrix of RGB pixel values sampled from the whole-slide.

    Notes
    -----
    If Dask is configured, it is used to distribute the computation.

    See Also
    --------
    histomicstk.preprocessing.color_normalization.reinhard

    """
    import dask
    import dask.distributed

    if (sample_fraction is None) == (sample_approximate_total is None):
        msg = 'Exactly one of sample_fraction and sample_approximate_total must have a value.'
        raise ValueError(msg)

    ts = large_image.getTileSource(slide_path, style=style)

    if magnification is None:
        magnification = ts.getMetadata()['magnification']

    # get entire whole-slide image at low resolution
    scale_lres = {'magnification': tissue_seg_mag}
    im_lres, _ = ts.getRegion(
        format=large_image.tilesource.TILE_FORMAT_NUMPY,
        scale=scale_lres,
        frame=frame,
    )

    # check number of channels
    if len(im_lres.shape) <= 2 or im_lres.shape[2] == 1:
        im_lres = np.dstack((im_lres, im_lres, im_lres))
        if default_img_inversion:
            invert_image = True
    else:
        im_lres = im_lres[:, :, :3]

    # perform image inversion
    if invert_image:
        im_lres = (np.iinfo(im_lres.dtype).max
                   if im_lres.dtype.kind == 'u'
                   else np.max(im_lres)) - im_lres

    # compute foreground mask of whole-slide image at low-res.
    # it will actually be a background mask if background is set.
    im_fgnd_mask_lres = bool(background) ^ simple_mask(im_lres)

    if sample_approximate_total is not None:
        scale_ratio = float(magnification) / tissue_seg_mag
        total_fgnd_pixels = np.count_nonzero(im_fgnd_mask_lres) * scale_ratio ** 2
        sample_fraction = sample_approximate_total / total_fgnd_pixels

    # broadcasting fgnd mask to all dask workers
    try:
        c = dask.distributed.get_client()

        [im_fgnd_mask_lres] = c.scatter([im_fgnd_mask_lres],
                                        broadcast=True)
    except ValueError:
        pass

    # generate sample pixels
    sample_pixels = []

    iter_args = dict(scale=dict(magnification=magnification),
                     format=large_image.tilesource.TILE_FORMAT_NUMPY)

    total_tiles = ts.getSingleTile(**iter_args)['iterator_range']['position']

    for position in range(0, total_tiles, tile_grouping):

        sample_pixels.append(dask.delayed(_sample_pixels_tile)(
            slide_path, iter_args,
            (position, min(tile_grouping, total_tiles - position)),
            sample_fraction, tissue_seg_mag, min_coverage,
            im_fgnd_mask_lres, invert_image=invert_image, style=style,
            default_img_inversion=default_img_inversion))

    # concatenate pixel values in list
    if sample_pixels:
        sample_pixels = (dask.delayed(np.concatenate)(sample_pixels, 0)
                         .compute())
    else:
        print('Sampling could not identify any foreground regions.')

    return sample_pixels


def _sample_pixels_tile(slide_path, iter_args, positions, sample_fraction,
                        tissue_seg_mag, min_coverage, im_fgnd_mask_lres,
                        invert_image=False, style=None, default_img_inversion=False):
    start_position, position_count = positions
    sample_pixels = [np.empty((0, 3))]
    ts = large_image.getTileSource(slide_path, style=style)
    for position in range(start_position, start_position + position_count):
        tile = ts.getSingleTile(tile_position=position, **iter_args)
        # get current region in base_pixels
        rgn_hres = {'left': tile['gx'], 'top': tile['gy'],
                    'right': tile['gx'] + tile['gwidth'],
                    'bottom': tile['gy'] + tile['gheight'],
                    'units': 'base_pixels'}

        # get foreground mask for current tile at low resolution
        rgn_lres = ts.convertRegionScale(rgn_hres,
                                         targetScale={'magnification':
                                                      tissue_seg_mag},
                                         targetUnits='mag_pixels')

        top = int(rgn_lres['top'])
        bottom = int(rgn_lres['bottom'])
        left = int(rgn_lres['left'])
        right = int(rgn_lres['right'])

        tile_fgnd_mask_lres = im_fgnd_mask_lres[top:bottom, left:right]

        # skip tile if there is not enough foreground in the slide
        cur_fgnd_frac = tile_fgnd_mask_lres.mean()

        if np.isnan(cur_fgnd_frac) or cur_fgnd_frac <= min_coverage:
            continue

        # check number of channels
        if len(tile['tile'].shape) <= 2 or tile['tile'].shape[2] == 1:
            im_tile = np.dstack((tile['tile'], tile['tile'], tile['tile']))
            if default_img_inversion:
                invert_image = True

        else:
            im_tile = tile['tile'][:, :, :3]

        # perform image inversion
        if invert_image:
            im_tile = (np.iinfo(im_tile.dtype).max
                       if im_tile.dtype.kind == 'u'
                       else np.max(im_tile)) - im_tile
        # get tile foreground mask at resolution of current tile
        tile_fgnd_mask = np.array(PIL.Image.fromarray(tile_fgnd_mask_lres).resize(
            im_tile.shape[:2],
            resample=PIL.Image.NEAREST,
        ))

        # generate linear indices of sample pixels in fgnd mask
        nz_ind = np.nonzero(tile_fgnd_mask.flatten())[0]

        # Handle fractions in the desired sample size by rounding up
        # or down, weighted by the fractional amount.
        float_samples = sample_fraction * nz_ind.size
        num_samples = int(np.floor(float_samples))
        num_samples += np.random.binomial(1, float_samples - num_samples)

        sample_ind = np.random.choice(nz_ind, num_samples)

        # convert rgb tile image to Nx3 array
        tile_pix_rgb = np.reshape(im_tile, (-1, 3))

        # add rgb triplet of sample pixels
        sample_pixels.append(tile_pix_rgb[sample_ind, :])

    return np.concatenate(sample_pixels, 0)



"""
This package contains implementations of state-of-th-art methods for
segmenting nuclei from histopathology images.
"""

from .detect_nuclei_kofahi import detect_nuclei_kofahi
from .detect_tile_nuclei import detect_tile_nuclei
from .gaussian_voting import gaussian_voting
from .gvf_tracking import gvf_tracking
from .max_clustering import max_clustering
from .min_model import min_model

__all__ = (
    'detect_nuclei_kofahi',
    'gaussian_voting',
    'gvf_tracking',
    'max_clustering',
    'min_model',
    'detect_tile_nuclei',
)



# import sub-packages to support nested calls
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _importlib_version

from . import segmentation  # must be imported before features
from . import utils  # must be imported before other packages
from . import (annotations_and_masks, features, filters, preprocessing,
               saliency, workflows)

try:
    __version__ = _importlib_version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass


# list out things that are available for public use
__all__ = (

    # sub-packages
    'features',
    'filters',
    'preprocessing',
    'segmentation',
    'utils',
    'annotations_and_masks',
    'saliency',
    'workflows',
)



import numpy as np

from ._trace_object_boundaries_cython import _trace_object_boundaries_cython


def trace_object_boundaries(im_label,
                            conn=4, trace_all=False,
                            x_start=None, y_start=None,
                            max_length=None,
                            simplify_colinear_spurs=True,
                            eps_colinear_area=0.01,
                            region_props=None):
    """Performs exterior boundary tracing of one or more objects in a label
    mask. If a starting point is not provided then a raster scan will be performed
    to identify the starting pixel.

    Parameters
    ----------
    im_label : array_like
        A binary mask image.
    conn : int
        Neighborhood connectivity to evaluate. Valid values are 4 or 8.
        Default value = 4.
    trace_all : bool
        Specify True if you want to trace boundaries of all objects.
        Default = False
    x_start : int
        Starting horizontal coordinate to begin tracing. Default value = None.
    y_start : int
        Starting vertical coordinate to begin tracing. Default value = None.
    max_length : int
        Maximum boundary length to trace before terminating. Default value =
        None.
    simplify_colinear_spurs : bool
        If True colinear streaks/spurs in the object boundary will be
        simplified/removed. Note that if the object boundary is entirely
        colinear then the object itself will be removed. Default = True
    eps_colinear_area : int
        Minimum area of triangle formed by three consecutive points on the
        contour for them to be considered as non-colinear. Default value =
        0.01.

    Notes
    -----
    The Improved Simple Boundary Follower (ISBF) from the reference below is
    used for 4-connected tracing. This algorithm provides accurate tracing with
    competitive execution times. 8-connected tracing is implemented using the
    Moore tracing algorithm.

    Returns
    -------
    X : array_like
        A set of 1D array of horizontal coordinates of contour seed pixels for
        tracing.
    Y : array_like
        A set of 1D array of the vertical coordinates of seed pixels for
        tracing.

    References
    ----------
    .. [#] J. Seo et al "Fast Contour-Tracing Algorithm Based on a Pixel-
       Following Method for Image Sensors" in Sensors,vol.16,no.353,
       doi:10.3390/s16030353, 2016.

    """
    from skimage.measure import regionprops

    if max_length is None:
        max_length = float('inf')

    X = []
    Y = []
    selected_rows = []

    if trace_all:

        rprops = region_props if region_props else regionprops(im_label)
        numLabels = len(rprops)

        x_start = -1
        y_start = -1

        for i in range(numLabels):

            # get bounds of label mask
            min_row, min_col, max_row, max_col = rprops[i].bbox

            # grab label mask
            lmask = (
                im_label[
                    min_row:max_row, min_col:max_col,
                ] == rprops[i].label
            ).astype(bool)

            mrows = max_row - min_row + 2
            mcols = max_col - min_col + 2

            mask = np.zeros((mrows, mcols))
            mask[1:mrows - 1, 1:mcols - 1] = lmask

            by, bx = _trace_object_boundaries_cython(
                np.ascontiguousarray(
                    mask, dtype=int), conn, x_start, y_start, max_length,
            )

            bx = bx + min_row - 1
            by = by + min_col - 1

            if simplify_colinear_spurs:
                bx, by = _remove_thin_colinear_spurs(bx, by,
                                                     eps_colinear_area)

            if len(bx) > 0:
                selected_rows.append(i)
                X.append(bx)
                Y.append(by)

    else:

        rprops = regionprops(im_label.astype(int))
        numLabels = len(rprops)

        if numLabels > 1:
            msg = 'Number of labels should be 1 !!'
            raise ValueError(msg)

        if (x_start is None and y_start is not None) | \
                (x_start is not None and y_start is None):
            msg = 'x_start or y_start is not defined !!'
            raise ValueError(msg)

        if x_start is None and y_start is None:
            x_start = -1
            y_start = -1

        by, bx = _trace_object_boundaries_cython(
            np.ascontiguousarray(
                im_label, dtype=int), conn, x_start, y_start, max_length,
        )

        if simplify_colinear_spurs:
            bx, by = _remove_thin_colinear_spurs(bx, by,
                                                 eps_colinear_area)

        if len(bx) > 0:
            X.append(bx)
            Y.append(by)
    if region_props:
        return X, Y, selected_rows

    return X, Y


def _remove_thin_colinear_spurs(px, py, eps_colinear_area=0):
    """Simplifies the given list of points by removing colinear spurs
    """
    keep = []  # indices of points to keep

    anchor = -1
    testpos = 0

    while testpos < len(px):

        # get coords of next triplet of points to test
        if testpos == len(px) - 1:
            if not keep:
                break
            nextpos = keep[0]
        else:
            nextpos = testpos + 1

        ind = [anchor, testpos, nextpos]
        x1, x2, x3 = px[ind]
        y1, y2, y3 = py[ind]

        # compute area of triangle formed by triplet
        area = 0.5 * np.linalg.det(
            np.array([[x1, x2, x3], [y1, y2, y3], [1, 1, 1]]),
        )

        # if area > cutoff, add testpos to keep and move anchor to testpos
        if abs(area) > eps_colinear_area:

            keep.append(testpos)  # add testpos to keep list
            anchor = testpos      # make testpos the next anchor point
            testpos += 1

        else:

            testpos += 1

    px = px[keep]
    py = py[keep]

    return px, py



"""
Created on Mon Aug 12 18:33:48 2019.

@author: tageldim

"""

import copy
import io
import os
from warnings import warn

import numpy as np
import pandas as pd
from imageio import imwrite
from matplotlib.patches import Polygon as mpPolygon
from PIL import Image
from shapely.geometry.polygon import Polygon

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    _get_and_add_element_to_roi, _get_element_mask, _get_idxs_for_all_rois,
    get_bboxes_from_slide_annotations,
    get_idxs_for_annots_overlapping_roi_by_bbox, get_image_from_htk_response,
    get_scale_factor_and_appendStr, scale_slide_annotations)
from histomicstk.annotations_and_masks.masks_to_annotations_handler import \
    get_contours_from_mask


def get_roi_mask(
        slide_annotations, element_infos, GTCodes_df,
        idx_for_roi, iou_thresh=0.0, roiinfo=None,
        crop_to_roi=True, use_shapely=True,
        verbose=False, monitorPrefix=''):
    """Parse annotations and gets a ground truth mask for a single ROI.

    This will look at all slide annotations and get ones that
    overlap with the region of interest (ROI) and assigns them to mask.

    Parameters
    ----------
    slide_annotations : list of dicts
        response from server request
    element_infos : pandas DataFrame.
        The columns annidx and elementidx
        encode the dict index of annotation document and element,
        respectively, in the original slide_annotations list of dictionaries.
        This can be obain by get_bboxes_from_slide_annotations() method
    GTCodes_df : pandas Dataframe
        the ground truth codes and information dataframe.
        WARNING: Modified inside this method so pass a copy.
        This is a dataframe that is indexed by the annotation group name and
        has the following columns:
        - group: group name of annotation (string), eg. mostly_tumor
        - overlay_order: int, how early to place the annotation in the
        mask. Larger values means this annotation group is overlaid
        last and overwrites whatever overlaps it.
        - GT_code: int, desired ground truth code (in the mask)
        Pixels of this value belong to corresponding group (class)
        - is_roi: Flag for whether this group encodes an ROI
        - is_background_class: Flag, whether this group is the default
        fill value inside the ROI. For example, you may decide that
        any pixel inside the ROI is considered stroma.
    idx_for_roi : int
        index of ROI within the element_infos dataframe.
    iou_thresh : float
        how much bounding box overlap is enough to
        consider an annotation to belong to the region of interest
    roiinfo : pandas series or dict
        contains information about the roi. Keys will be added to this
        index containing info about the roi like bounding box
        location and size.
    crop_to_roi : bool
        flag of whether to crop polygons to roi
        (prevent overflow beyond roi edge)
    use_shapely : bool
        flag of whether to precisely determine whether an element
        belongs to an ROI using shapely polygons. Slightly slower. If
        set to False, overlapping bounding box is used as a cheap but
        less precise indicator of inclusion.
    verbose : bool
        Print progress to screen?
    monitorPrefix : str
        text to prepend to printed statements

    Returns
    -------
    Np array
        (N x 2), where pixel values encode class membership.
        IMPORTANT NOTE: Zero pixels have special meaning and do NOT
        encode specific ground truth class. Instead, they simply
        mean Outside ROI and should be IGNORED during model training
        or evaluation.
    Dict
        information about ROI

    """
    # This stores information about the ROI like bounds, slide_name, etc
    # Allows passing many parameters and good forward/backward compatibility
    if roiinfo is None:
        roiinfo = {}

    # isolate annotations that potentially overlap (belong to) mask (incl. ROI)
    overlaps = get_idxs_for_annots_overlapping_roi_by_bbox(
        element_infos, idx_for_roi=idx_for_roi, iou_thresh=iou_thresh)
    idxs_for_all_rois = _get_idxs_for_all_rois(
        GTCodes=GTCodes_df, element_infos=element_infos)
    overlaps = list(set(overlaps) - set(idxs_for_all_rois))
    elinfos_roi = element_infos.loc[[idx_for_roi] + overlaps, :]

    # Add roiinfo
    roiinfo['XMIN'] = int(np.min(elinfos_roi.xmin))
    roiinfo['YMIN'] = int(np.min(elinfos_roi.ymin))
    roiinfo['XMAX'] = int(np.max(elinfos_roi.xmax))
    roiinfo['YMAX'] = int(np.max(elinfos_roi.ymax))
    roiinfo['BBOX_WIDTH'] = roiinfo['XMAX'] - roiinfo['XMIN']
    roiinfo['BBOX_HEIGHT'] = roiinfo['YMAX'] - roiinfo['YMIN']

    # get roi polygon
    if use_shapely:
        coords, _ = _get_element_mask(
            elinfo=elinfos_roi.loc[idx_for_roi],
            slide_annotations=slide_annotations)
        roi_polygon = Polygon(coords)

    # Init mask
    ROI = np.zeros(
        (roiinfo['BBOX_HEIGHT'], roiinfo['BBOX_WIDTH']), dtype=np.uint8)

    # only parse if roi is polygonal or rectangular
    if elinfos_roi.loc[idx_for_roi, 'type'] == 'point':
        msg = 'roi cannot be a point!'
        raise Exception(msg)

    # make sure ROI is overlaid first & assigned background class if relevant
    roi_group = elinfos_roi.loc[idx_for_roi, 'group']
    GTCodes_df.loc[roi_group, 'overlay_order'] = np.min(
        GTCodes_df.loc[:, 'overlay_order']) - 1
    bck_classes = GTCodes_df.loc[
        GTCodes_df.loc[:, 'is_background_class'] == 1, :]
    if bck_classes.shape[0] > 0:
        GTCodes_df.loc[
            roi_group, 'GT_code'] = bck_classes.iloc[0, :]['GT_code']

    # Add annotations in overlay order
    overlay_orders = sorted(set(GTCodes_df.loc[:, 'overlay_order']))
    N_elements = elinfos_roi.shape[0]
    elNo = 0
    for overlay_level in overlay_orders:

        # get indices of relevant groups
        relevant_groups = list(GTCodes_df.loc[
            GTCodes_df.loc[:, 'overlay_order'] == overlay_level, 'group'])
        relIdxs = []
        for group_name in relevant_groups:
            relIdxs.extend(list(elinfos_roi.loc[
                elinfos_roi.group == group_name, :].index))

        # get relevnt infos and sort from largest to smallest (by bbox area)
        # so that the smaller elements are layered last. This helps partially
        # address issues describe in:
        # https://github.com/DigitalSlideArchive/HistomicsTK/issues/675
        elinfos_relevant = elinfos_roi.loc[relIdxs, :].copy()
        elinfos_relevant.sort_values(
            'bbox_area', axis=0, ascending=False, inplace=True)

        # Go through elements and add to ROI mask
        for _elId, elinfo in elinfos_relevant.iterrows():

            elNo += 1
            elcountStr = '%s: Overlay level %d: Element %d of %d: %s' % (
                monitorPrefix, overlay_level, elNo, N_elements,
                elinfo['group'])
            if verbose:
                print(elcountStr)

            # now add element to ROI
            ROI = _get_and_add_element_to_roi(
                elinfo=elinfo, slide_annotations=slide_annotations, ROI=ROI,
                roiinfo=roiinfo, roi_polygon=roi_polygon,
                GT_code=GTCodes_df.loc[elinfo['group'], 'GT_code'],
                use_shapely=use_shapely, verbose=verbose,
                monitorPrefix=elcountStr)

            # save a copy of ROI-only mask to crop to it later if needed
            if crop_to_roi and (overlay_level == GTCodes_df.loc[
                    roi_group, 'overlay_order']):
                roi_only_mask = ROI.copy()

    # Crop polygons to roi if needed (prevent 'overflow' beyond roi edge)
    if crop_to_roi:
        ROI[roi_only_mask == 0] = 0

    # tighten boundary --remember, so far we've use element bboxes to
    # make an over-estimated margin around ROI boundary.
    nz = np.nonzero(ROI)
    ymin, xmin = (np.min(arr) for arr in nz)
    ymax, xmax = (np.max(arr) for arr in nz)
    ROI = ROI[ymin:ymax, xmin:xmax]

    # update roi offset
    roiinfo['XMIN'] += xmin
    roiinfo['YMIN'] += ymin
    roiinfo['XMAX'] += xmin
    roiinfo['YMAX'] += ymin
    roiinfo['BBOX_WIDTH'] = roiinfo['XMAX'] - roiinfo['XMIN']
    roiinfo['BBOX_HEIGHT'] = roiinfo['YMAX'] - roiinfo['YMIN']

    return ROI, roiinfo


def get_mask_from_slide(
        GTCodes_dict, roiinfo, slide_annotations,
        element_infos, sf=1.0, get_roi_mask_kwargs=None):
    """Parse region from the slide and get its corresponding labeled mask.

    This is a wrapper around get_roi_mask() which should be referred to for
    implementation details. If roiinfo is None, all annotations in the slide
    are parsed into labeled image (mask) form. Otherwise, the bounding box
    coordinates in roiinfo are used.

    Parameters
    ----------
    GTCodes_dict : dict
        the ground truth codes and information dict.
        This is a dict that is indexed by the annotation group name and
        each entry is in turn a dict with the following keys:
        - group: group name of annotation (string), eg. mostly_tumor
        - overlay_order: int, how early to place the annotation in the
        mask. Larger values means this annotation group is overlaid
        last and overwrites whatever overlaps it.
        - GT_code: int, desired ground truth code (in the mask)
        Pixels of this value belong to corresponding group (class)
        - is_roi: Flag for whether this group encodes an ROI
        - is_background_class: Flag, whether this group is the default
        fill value inside the ROI. For example, you may decide that
        any pixel inside the ROI is considered stroma.

    roiinfo : dict or None
        if not None, has keys 'XMIN', 'XMAX', 'YMIN', 'YMAX' for slide
        region coordinates (AT BASE MAGNIFICATION) to get labeled image
        (mask) for.

    sf : float
        scale factor to multiple coordinates (eg 0.5 would halve size)

    slide_annotations : list
        Make sure you have used
        scale_slide_annotations() to scale them up/down by sf BEFOREHAND.

    element_infos : pandas DataFrame.
        The columns annidx and elementidx
        encode the dict index of annotation document and element,
        respectively, in the original slide_annotations list of dictionaries.
        This can be obained by get_bboxes_from_slide_annotations() method.
        Make sure you have used scale_slide_annotations().

    get_roi_mask_kwargs : dict
        extra kwargs for get_roi_mask()

    Returns
    -------
    Np array
        (N x 2), where pixel values encode class membership.
        IMPORTANT NOTE: Zero pixels have special meaning and do NOT
        encode specific ground truth class. Instead, they simply
        mean Outside mask and should be IGNORED during model training
        or evaluation.

    Dict
        information about mask

    """
    from pandas import DataFrame

    # convert from dict to required dataframe
    if get_roi_mask_kwargs is None:
        get_roi_mask_kwargs = {}
    GTCodes = DataFrame.from_dict(GTCodes_dict, orient='index')

    # some sanity checks
    assert all(j in GTCodes.columns for j in [
        'group', 'overlay_order', 'GT_code', 'is_roi', 'is_background_class',
        'color']), 'GTCodes_dict does not follow schema'
    assert all(GTCodes.loc[:, 'GT_code'] > 0), 'All GT_code must be > 0'
    assert sf > 0, 'sf must be positive.'
    assert (roiinfo['XMAX'] > roiinfo['XMIN'])
    assert (
        roiinfo['YMAX'] > roiinfo['YMIN'])

    # use given ROI bounds, after scaling
    XMIN = int(roiinfo['XMIN'] * sf)
    YMIN = int(roiinfo['YMIN'] * sf)
    XMAX = int(roiinfo['XMAX'] * sf)
    YMAX = int(roiinfo['YMAX'] * sf)
    WIDTH = XMAX - XMIN
    HEIGHT = YMAX - YMIN

    # add to slide annotations list
    slide_annotations.append({'annotation': {
        'description': '',
        'elements': [
            {'center': [int(XMIN + WIDTH / 2), int(YMIN + HEIGHT / 2), 0],
             'width': WIDTH,
             'height': HEIGHT,
             'normal': [0, 0, 1],
             'rotation': 0,
             'group': 'super_roi',
             'label': {'value': 'super_roi'},
             'lineColor': 'rgb(0, 0, 0)',
             'fillColor': 'rgba(0, 0, 0, 0)',
             'lineWidth': 4.6,
             'type': 'rectangle'},
        ],
        'name': 'superROI'},
    })

    # add to bounding boxes dataframe
    element_infos = pd.concat([element_infos, pd.DataFrame([{
        'annidx': len(slide_annotations) - 1,
        'elementidx': 0,
        'type': 'rectangle',
        'group': 'super_roi',
        'xmin': XMIN,
        'xmax': XMAX,
        'ymin': YMIN,
        'ymax': YMAX,
        'bbox_area': WIDTH * HEIGHT,
    }])], ignore_index=True)

    # find roi and background codes to use later
    roi_codes = list(GTCodes.loc[GTCodes.loc[:, 'is_roi'] == 1, 'GT_code'])
    bck_code = GTCodes.loc[
        GTCodes.loc[:, 'is_background_class'] == 1, 'GT_code']
    if bck_code.shape[0] > 0:
        bck_code = int(bck_code.iloc[0])
    else:
        bck_code = 0

    # add to gtcodes dataframe
    assert np.max(GTCodes.loc[:, 'GT_code']) < 255
    GTCodes.loc[:, 'is_roi'] = 0  # treat other ROIs as ordinary annotations
    GTCodes.loc[:, 'is_background_class'] = 0  # we'll adjust later
    GTCodes = pd.concat([GTCodes, pd.DataFrame([{
        'GT_code': 255,
        'overlay_order': 0,
        'color': 'rgb(0,0,0)',
        'group': 'super_roi',
        'is_background_class': 0,
        'is_roi': 1,
    }])], ignore_index=True)
    GTCodes.index = GTCodes.loc[:, 'group']

    # now get mask
    ROI, roiinfo = get_roi_mask(
        slide_annotations=slide_annotations, element_infos=element_infos,
        GTCodes_df=GTCodes.copy(),
        idx_for_roi=element_infos.index[-1],  # <- bounding roi
        **get_roi_mask_kwargs)
    ROI[ROI == 255] = 0

    # replace roi codes with background code
    for roi_code in roi_codes:
        ROI[ROI == roi_code] = bck_code

    # scale back coords
    roiinfo = {k: int(v / sf) for k, v in roiinfo.items()}

    return ROI, roiinfo


def _visualize_annotations_on_rgb(
        rgb, contours_list, linewidth=0.2, x_offset=0, y_offset=0,
        text=False):
    import matplotlib.pyplot as plt

    # later on flipped by matplotlib for weird reason
    rgb = np.flipud(rgb)

    fig = plt.figure(
        figsize=(rgb.shape[1] / 1000, rgb.shape[0] / 1000), dpi=100)
    ax = plt.subplot(111)
    ax.imshow(rgb)

    plt.axis('off')
    ax = plt.gca()
    ax.set_xlim(0.0, rgb.shape[1])
    ax.set_ylim(0.0, rgb.shape[0])

    for _idx, ann in enumerate(contours_list):
        xy = np.array([
            [int(j) for j in ann[k].split(',')]
            for k in ('coords_x', 'coords_y')]).T
        xy[:, 0] = xy[:, 0] - x_offset
        xy[:, 1] = rgb.shape[0] - (xy[:, 1] - y_offset) + 1
        polygon = mpPolygon(
            xy=xy,
            color=[int(j) / 255 for j in ann['color'].split(
                'rgb(')[1][:-1].split(',')],
            closed=True, fill=False,
            linewidth=linewidth,
        )
        ax.add_patch(polygon)

        # add label text
        if text:
            txtshift = 0
            size = 1e-4 * rgb.shape[1]
            ax.text(
                int(np.min(xy[:, 0])),
                int(np.max(xy[:, 1])) - txtshift,
                ann['group'][:5],
                color='w', fontsize=size, backgroundcolor='none',
            )

    ax.axis('off')
    fig.subplots_adjust(bottom=0, top=1, left=0, right=1)

    buf = io.BytesIO()
    plt.savefig(buf, format='png', pad_inches=0, dpi=1000)
    buf.seek(0)
    rgb_vis = np.uint8(Image.open(buf))[..., :3]
    plt.close()

    return rgb_vis


def _sanity_checks(
        MPP, MAG, mode, bounds, idx_for_roi, get_roi_mask_kwargs,
        get_rgb, get_contours, get_visualization):

    # MPP precedes MAG
    if all(j is not None for j in (MPP, MAG)):
        MAG = None

    # some sanity checks

    for mf in (MPP, MAG):
        if mf is not None:
            assert mf > 0, 'MPP or MAG must be positive.'

    if mode in ['wsi', 'min_bounding_box']:
        bounds = None
        idx_for_roi = None

    if idx_for_roi is not None:
        mode = 'polygonal_bounds'
    elif bounds is not None:
        mode = 'manual_bounds'

    assert mode in [
        'wsi', 'min_bounding_box', 'manual_bounds', 'polygonal_bounds'], \
        'mode %s not recognized' % mode

    if get_visualization:
        get_contours = True
        assert get_rgb, 'cannot get visualization without rgb.'

    if not get_roi_mask_kwargs['crop_to_roi']:
        assert not get_rgb, \
            'Handling overflowing annotations while also getting RGB is not currently supported.'
        assert not get_visualization, \
            'Handling overflowing annotations while also getting RGB is not currently supported.'

    return (
        MPP, MAG, mode, bounds, idx_for_roi, get_roi_mask_kwargs,
        get_rgb, get_contours, get_visualization)


def _get_roi_bounds_by_run_mode(
        gc, slide_id, mode, bounds, element_infos, idx_for_roi, sf):

    if mode == 'polygonal_bounds':
        # get bounds based on specified polygonal/rotated roi
        elinfo = element_infos.loc[idx_for_roi]
        bounds = {
            'XMIN': int(elinfo['xmin'] / sf),
            'XMAX': int(elinfo['xmax'] / sf),
            'YMIN': int(elinfo['ymin'] / sf),
            'YMAX': int(elinfo['ymax'] / sf),
        }

    elif mode == 'manual_bounds':
        assert (bounds['XMAX'] > bounds['XMIN'])
        assert (
            bounds['YMAX'] > bounds['YMIN'])

    elif mode == 'min_bounding_box':
        # get minimum box for all annotations in slide
        bounds = {
            'XMIN': int(np.min(element_infos.xmin) / sf),
            'YMIN': int(np.min(element_infos.ymin) / sf),
            'XMAX': int(np.max(element_infos.xmax) / sf),
            'YMAX': int(np.max(element_infos.ymax) / sf),
        }
    else:
        # get scaled up/down version of mask of whole slide
        slide_info = gc.get('/item/%s/tiles' % slide_id)
        bounds = {
            'XMIN': 0,
            'XMAX': slide_info['sizeX'],
            'YMIN': 0,
            'YMAX': slide_info['sizeY'],
        }

    return bounds


def _get_rgb_and_pad_roi(gc, slide_id, bounds, appendStr, ROI, tau=10):

    getStr = \
        '/item/%s/tiles/region?left=%d&right=%d&top=%d&bottom=%d&encoding=PNG' \
        % (slide_id,
           bounds['XMIN'], bounds['XMAX'],
           bounds['YMIN'], bounds['YMAX'])
    getStr += appendStr
    resp = gc.get(getStr, jsonResp=False)
    rgb = get_image_from_htk_response(resp)

    # sometimes there's a couple of pixel difference d.t. rounding, so pad
    pad_y = rgb.shape[0] - ROI.shape[0]
    pad_x = rgb.shape[1] - ROI.shape[1]
    assert all(np.abs(j) < tau for j in (pad_y, pad_x)), \
        'too much difference in size between image and mask.  Something is wrong!'

    if pad_y > 0:
        ROI = np.pad(ROI, pad_width=((0, pad_y), (0, 0)), mode='constant')
    elif pad_y < 0:
        ROI = ROI[:pad_y, :]

    if pad_x > 0:
        ROI = np.pad(ROI, pad_width=((0, 0), (0, pad_x)), mode='constant')
    elif pad_x < 0:
        ROI = ROI[:, :pad_x]

    return rgb, ROI


def get_image_and_mask_from_slide(
        gc, slide_id, GTCodes_dict,
        MPP=5.0, MAG=None, mode='min_bounding_box',
        bounds=None, idx_for_roi=None,
        slide_annotations=None, element_infos=None,
        get_roi_mask_kwargs=None, get_contours_kwargs=None, linewidth=0.2,
        get_rgb=True, get_contours=True, get_visualization=True, tau=10):
    """Parse region from the slide and get its corresponding labeled mask.

    This is a wrapper around get_roi_mask() which should be referred to for
    implementation details.

    Parameters
    ----------
    gc : object
        girder client object to make requests, for example:
        gc = girder_client.GirderClient(apiUrl = APIURL)
        gc.authenticate(interactive=True)

    slide_id : str
        girder id for item (slide)

    GTCodes_dict : dict
        the ground truth codes and information dict.
        This is a dict that is indexed by the annotation group name and
        each entry is in turn a dict with the following keys:
        - group: group name of annotation (string), eg. mostly_tumor
        - overlay_order: int, how early to place the annotation in the
        mask. Larger values means this annotation group is overlaid
        last and overwrites whatever overlaps it.
        - GT_code: int, desired ground truth code (in the mask)
        Pixels of this value belong to corresponding group (class)
        - is_roi: Flag for whether this group encodes an ROI
        - is_background_class: Flag, whether this group is the default
        fill value inside the ROI. For example, you may decide that
        any pixel inside the ROI is considered stroma.

    MPP : float or None
        Microns-per-pixel -- best use this as it's more well-defined than
        magnification which is more scanner/manufacturer specific.
        MPP of 0.25 often roughly translates to 40x

    MAG : float or None
        If you prefer to use whatever magnification is reported in slide.
        If neither MPP or MAG is provided, everything is retrieved without
        scaling at base (scan) magnification.

    mode : str
        This specifies which part of the slide to get the mask from. Allowed
        modes include the following
        - wsi: get scaled up/down version of mask of whole slide
        - min_bounding_box: get minimum box for all annotations in slide
        - manual_bounds: use given ROI bounds provided by the 'bounds' param
        - polygonal_bounds: use the idx_for_roi param to get coordinates

    bounds : dict or None
        if not None, has keys 'XMIN', 'XMAX', 'YMIN', 'YMAX' for slide
        region coordinates (AT BASE MAGNIFICATION) to get labeled image
        (mask) for. Use this with the 'manual_bounds' run mode.

    idx_for_roi : int
        index of ROI within the element_infos dataframe.
        Use this with the 'polygonal_bounds' run mode.

    slide_annotations : list or None
        Give this parameter to avoid re-getting slide annotations. If you do
        provide the annotations, though, make sure you have used
        scale_slide_annotations() to scale them up/down by sf BEFOREHAND.

    element_infos : pandas DataFrame.
        The columns annidx and elementidx
        encode the dict index of annotation document and element,
        respectively, in the original slide_annotations list of dictionaries.
        This can be obained by get_bboxes_from_slide_annotations() method.
        Make sure you have used scale_slide_annotations().

    get_roi_mask_kwargs : dict
        extra kwargs for get_roi_mask()

    get_contours_kwargs : dict
        extra kwargs for get_contours_from_mask()

    linewidth : float
        visualization line width

    get_rgb: bool
        get rgb image?

    get_contours : bool
        get annotation contours? (relative to final mask)

    get_visualization : bool
        get overlaid annotation bounds over RGB for visualization

    tau : int
        maximum difference (in pixels) between fetched image and mask allowed.
        Above this threshold, an error is raised indicating you may have some
        problem in your parameters or elsewhere. If the difference is less then
        tau, the rgb image and mask are resized to match each other before
        being returned

    Returns
    -------
    dict
        Results dict containing one or more of the following keys
        bounds: dict of bounds at scan magnification
        ROI - (mxn) labeled image (mask)
        rgb - (mxnx3 np array) corresponding rgb image
        contours - list, each entry is a dict version of a row from the output
        of masks_to_annotations_handler.get_contours_from_mask()
        visualization - (mxnx3 np array) visualization overlay

    """
    from pandas import DataFrame

    get_roi_mask_kwargs = get_roi_mask_kwargs or {}
    get_contours_kwargs = get_contours_kwargs or {}
    # important sanity checks
    (MPP, MAG, mode, bounds, idx_for_roi, get_roi_mask_kwargs,
     get_rgb, get_contours, get_visualization) = _sanity_checks(
        MPP, MAG, mode, bounds, idx_for_roi, get_roi_mask_kwargs,
        get_rgb, get_contours, get_visualization)

    # calculate the scale factor
    sf, appendStr = get_scale_factor_and_appendStr(
        gc=gc, slide_id=slide_id, MPP=MPP, MAG=MAG)

    if slide_annotations is not None:
        assert element_infos is not None, 'must also provide element_infos'
    else:
        # get annotations for slide
        slide_annotations = gc.get('/annotation/item/' + slide_id)

        # scale up/down annotations by a factor
        slide_annotations = scale_slide_annotations(slide_annotations, sf=sf)

        # get bounding box information for all annotations -> scaled by sf
        element_infos = get_bboxes_from_slide_annotations(slide_annotations)

    # Determine get region based on run mode, keeping in mind that it
    # must be at BASE MAGNIFICATION coordinates before it is passed
    # on to get_mask_from_slide()
    bounds = _get_roi_bounds_by_run_mode(
        gc=gc, slide_id=slide_id, mode=mode, bounds=bounds,
        element_infos=element_infos, idx_for_roi=idx_for_roi, sf=sf)
    result = {'bounds': bounds}

    # get mask for specified area
    if mode == 'polygonal_bounds':
        # get roi mask and info
        ROI, _ = get_roi_mask(
            slide_annotations=slide_annotations, element_infos=element_infos,
            GTCodes_df=DataFrame.from_dict(GTCodes_dict, orient='index'),
            idx_for_roi=idx_for_roi, **get_roi_mask_kwargs)
    else:
        ROI, _ = get_mask_from_slide(
            GTCodes_dict=GTCodes_dict, roiinfo=copy.deepcopy(bounds),
            slide_annotations=slide_annotations, element_infos=element_infos,
            sf=sf, get_roi_mask_kwargs=get_roi_mask_kwargs)

    # get RGB
    if get_rgb:
        rgb, ROI = _get_rgb_and_pad_roi(
            gc=gc, slide_id=slide_id, bounds=bounds,
            appendStr=appendStr, ROI=ROI, tau=tau)
        result['rgb'] = rgb

    # pack result (we have to do it here in case of padding)
    result['ROI'] = ROI

    # get contours
    if get_contours:
        contours_list = get_contours_from_mask(
            MASK=ROI,
            GTCodes_df=DataFrame.from_dict(GTCodes_dict, orient='index'),
            **get_contours_kwargs)
        contours_list = contours_list.to_dict(orient='records')
        result['contours'] = contours_list

    # get visualization of annotations on RGB
    if get_visualization:
        result['visualization'] = _visualize_annotations_on_rgb(
            rgb=rgb, contours_list=contours_list, linewidth=linewidth)

    return result


def _roi_getter_asis(
        gc, slide_id, GTCodes_dict, slide_annotations, element_infos,
        get_kwargs, monitor='', verbose=False):
    """Download special ROI regions as-is, even if they are very large."""
    from pandas import DataFrame

    # get idx of all 'special' roi annotations
    GTCodes_df = DataFrame.from_dict(GTCodes_dict, orient='index')
    idxs_for_all_rois = _get_idxs_for_all_rois(
        GTCodes=GTCodes_df, element_infos=element_infos)

    # go through rois and download as-is
    for roino, idx_for_roi in enumerate(idxs_for_all_rois):

        roistr = '%s: roi %d of %d' % (
            monitor, roino + 1, len(idxs_for_all_rois))
        if verbose:
            print(roistr)

        try:
            roi_out = get_image_and_mask_from_slide(
                gc=gc, slide_id=slide_id, GTCodes_dict=GTCodes_dict,
                mode='polygonal_bounds', idx_for_roi=idx_for_roi,
                slide_annotations=slide_annotations,
                element_infos=element_infos, **get_kwargs)
        except Exception as e:
            problem = '\n   '
            problem += e.__repr__()
            problem += '\n'
            warn(problem)  # noqa B028
            roi_out = None

        yield roi_out


def _roi_getter_tiled(
        gc, slide_id, GTCodes_dict, slide_annotations, element_infos,
        sf, max_roiside,
        get_kwargs, monitor='', verbose=False):
    """Download special ROI regions in a tiled fashion."""
    # isolate rois
    rois = element_infos.loc[element_infos.loc[:, 'group'] == 'roi', :].copy()

    # split ROIs into max_roiside tiled regions
    for roidx, roi in rois.iterrows():

        # bounds for tiled sub-rois
        xbounds = list(np.arange(roi['xmin'], roi['xmax'], max_roiside))
        xbounds.append(roi['xmax'])
        ybounds = list(np.arange(roi['ymin'], roi['ymax'], max_roiside))
        ybounds.append(roi['ymax'])

        roidx += 1
        roistr = f'{monitor}: roi {roidx} of {rois.shape[0]}'
        if verbose:
            print(roistr)

        subroidx = 0
        nsubrois = (len(xbounds) - 1) * (len(ybounds) - 1)

        # go through tiled sub-rois
        for xi, xmin in enumerate(xbounds[:-1]):
            xmax = xbounds[xi + 1]
            for yi, ymin in enumerate(ybounds[:-1]):
                ymax = ybounds[yi + 1]

                subroidx += 1
                subroistr = f'{roistr}: sub-roi {subroidx} of {nsubrois}'
                if verbose:
                    print(subroistr)

                # get specified region
                get_kwargs['bounds'] = {
                    'XMIN': xmin / sf,
                    'XMAX': xmax / sf,
                    'YMIN': ymin / sf,
                    'YMAX': ymax / sf,
                }
                try:
                    roi_out = get_image_and_mask_from_slide(
                        gc=gc, slide_id=slide_id, GTCodes_dict=GTCodes_dict,
                        mode='manual_bounds',
                        slide_annotations=slide_annotations,
                        element_infos=element_infos,
                        **get_kwargs)
                except Exception as e:
                    problem = '\n'
                    problem += e.__repr__()
                    problem += '\n'
                    warn(problem)  # noqa B028
                    roi_out = None

                yield roi_out


def get_all_rois_from_slide(  # noqa: C901
        gc, slide_id, GTCodes_dict, save_directories,
        get_image_and_mask_from_slide_kwargs=None, max_roiside=None,
        slide_name=None, verbose=True, monitorPrefix=''):
    """Parse annotations and saves ground truth masks for ALL ROIs.

    Get all ROIs in a single slide. This is mainly uses
    get_image_and_mask_from_slide(), which should be referred to
    for implementation details.

    Parameters
    ----------
    gc : object
        girder client object to make requests, for example:
        gc = girder_client.GirderClient(apiUrl = APIURL)
        gc.authenticate(interactive=True)

    slide_id : str
        girder id for item (slide)

    GTCodes_dict : dict
        the ground truth codes and information dict.
        This is a dict that is indexed by the annotation group name and
        each entry is in turn a dict with the following keys:
        - group: group name of annotation (string), eg. mostly_tumor
        - overlay_order: int, how early to place the annotation in the
        mask. Larger values means this annotation group is overlaid
        last and overwrites whatever overlaps it.
        - GT_code: int, desired ground truth code (in the mask)
        Pixels of this value belong to corresponding group (class)
        - is_roi: Flag for whether this group encodes an ROI
        - is_background_class: Flag, whether this group is the default
        fill value inside the ROI. For example, you may decide that
        any pixel inside the ROI is considered stroma.

    save_directories : dict
        paths to directories to save data. Each entry is a string, and the
        following keys are allowed
        - ROI: path to save masks (labeled images)
        - rgb: path to save rgb images
        - contours: path to save annotation contours
        - visualization: path to save rgb visualization overlays

    get_image_and_mask_from_slide_kwargs : dict
        kwargs to pass to get_image_and_mask_from_slide()
        default values are assigned if speceific parameters are not given.

    max_roiside : int or None
        If int, this is the maximum allowed side for a downloaded region. If
        a region-of-interest is larger than this size, then it is tiled into
        non-overlapping regions whose maximal side is max_roiside.
        If None, the ROI is downloaded as-is, even if it was extremely large.
        If you know your slides have very large ROI annotations, the safer
        option is to set a max_roiside. A good value may be 5000-8000 pixels.

    slide_name : str or None
        If not given, it's inferred using a server request using girder client.

    verbose : bool
        Print progress to screen?

    monitorPrefix : str
        text to prepend to printed statements

    Returns
    -------
    list of dicts
        each entry contains the following keys
        - ROI: path to saved mask (labeled image)
        - rgb: path to saved rgb image
        - contours: path to saved annotation contours
        - visualization: path to saved rgb visualization overlay

    """
    from pandas import DataFrame

    # assign defaults if nothing given
    default_keyvalues = {
        'MPP': 5.0,
        'MAG': None,
        'get_roi_mask_kwargs': {
            'iou_thresh': 0.0, 'crop_to_roi': True,
            'use_shapely': True, 'verbose': False},
        'get_contours_kwargs': {
            'groups_to_get': None,
            'roi_group': 'roi',
            'get_roi_contour': True,
            'discard_nonenclosed_background': True,
            'background_group': 'mostly_stroma',
            'MIN_SIZE': 10, 'MAX_SIZE': None,
            'verbose': False, 'monitorPrefix': '',
        },
        'get_rgb': True,
        'get_contours': True,
        'get_visualization': True,
    }

    kvp = get_image_and_mask_from_slide_kwargs or {}  # for easy referencing
    for k, v in default_keyvalues.items():
        if k not in kvp.keys():
            kvp[k] = v

    # convert to df and sanity check
    GTCodes_df = DataFrame.from_dict(GTCodes_dict, orient='index')
    if any(GTCodes_df.loc[:, 'GT_code'] <= 0):
        msg = 'All GT_code must be > 0'
        raise Exception(msg)

    # if not given, assign name of first file associated with girder item
    if slide_name is None:
        resp = gc.get('/item/%s/files' % slide_id)
        slide_name = resp[0]['name']
        slide_name = slide_name[:slide_name.rfind('.')]

    # get annotations for slide
    slide_annotations = gc.get('/annotation/item/' + slide_id)

    # scale up/down annotations by a factor
    sf, _ = get_scale_factor_and_appendStr(
        gc=gc, slide_id=slide_id, MPP=kvp['MPP'], MAG=kvp['MAG'])
    slide_annotations = scale_slide_annotations(slide_annotations, sf=sf)

    # get bounding box information for all annotations
    element_infos = get_bboxes_from_slide_annotations(slide_annotations)

    # define roi_getter, which yields one roi at a time
    if max_roiside is None:
        roig = _roi_getter_asis(
            gc=gc, slide_id=slide_id, GTCodes_dict=GTCodes_dict,
            slide_annotations=slide_annotations, element_infos=element_infos,
            get_kwargs=kvp, monitor=monitorPrefix, verbose=verbose,
        )
    else:
        roig = _roi_getter_tiled(
            gc=gc, slide_id=slide_id, GTCodes_dict=GTCodes_dict,
            slide_annotations=slide_annotations, element_infos=element_infos,
            sf=sf, max_roiside=max_roiside,
            get_kwargs=kvp, monitor=monitorPrefix, verbose=verbose,
        )

    savenames = []

    for roi_out in roig:

        # if something went wrong, just move on
        if roi_out is None:
            continue

        # now save roi (mask, rgb, contours, vis)

        this_roi_savenames = {}
        ROINAMESTR = '%s_left-%d_top-%d_bottom-%d_right-%d' % (
            slide_name,
            roi_out['bounds']['XMIN'], roi_out['bounds']['YMIN'],
            roi_out['bounds']['YMAX'], roi_out['bounds']['XMAX'])

        for imtype in ['ROI', 'rgb', 'visualization']:
            if imtype in roi_out.keys():
                savename = os.path.join(
                    save_directories[imtype], ROINAMESTR + '.png')
                if verbose:
                    print('   Saving %s\n' % savename)
                imwrite(im=roi_out[imtype], uri=savename)
                this_roi_savenames[imtype] = savename

        if 'contours' in roi_out.keys():
            savename = os.path.join(
                save_directories['contours'], ROINAMESTR + '.csv')
            if verbose:
                print('   Saving %s\n' % savename)
            contours_df = DataFrame(roi_out['contours'])
            contours_df.to_csv(savename)
            this_roi_savenames['contours'] = savename

        savenames.append(this_roi_savenames)

    return savenames



"""
Created on Thu Dec 12 13:19:18 2019

@author: tageldim
"""
import json
import os

from histomicstk.workflows.workflow_runner import (Annotation_iterator,
                                                   Slide_iterator,
                                                   Workflow_runner)

# import warnings
# warnings.simplefilter('once', UserWarning)


def connect_to_api(apiurl, apikey=None, interactive=True):
    """Connect to a specific girder API.

    Parameters
    ----------
    apiurl : str
        URL to the AP to connect to.
    apikey : str
        API authentication token key
    interactive : bool
        Whether to use interactive mode instead.

    Returns
    -------
    girder_client.GirderClient
        Authenticated girder client.

    """
    import girder_client

    assert interactive or (apikey is not None)
    gc = girder_client.GirderClient(apiUrl=apiurl)
    if apikey is not None:
        interactive = False
    if interactive:
        gc.authenticate(interactive=True)
    else:
        gc.authenticate(apiKey=apikey)
    return gc


def get_absolute_girder_folderpath(gc, folder_id=None, folder_info=None):
    """Get absolute path for a girder folder.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    folder_id : str
        girder id of folder
    folder_info : dict
        folder info from the girder server

    Returns
    -------
    str
        absolute path to folder in the girder server.

    """
    assert any(j is not None for j in (folder_id, folder_info))
    if folder_id is not None:
        folder_info = gc.get('/folder/%s' % folder_id)
    fpath = gc.get('/folder/%s/rootpath' % folder_info['_id'])
    fpath = '/'.join([
        j['object']['name'] for j in fpath
        if j['object']['_modelType'] == 'folder'
    ]) + '/' + folder_info['name'] + '/'
    return fpath


def update_permissions_for_annotation(
        gc, annotation_id=None, annotation=None,
        groups_to_add=None, replace_original_groups=True,
        users_to_add=None, replace_original_users=True):
    """Update permissions for a single annotation.

    Parameters
    ----------
    gc : gider_client.GirderClient
        authenticated girder client instance
    annotation_id : str
        girder id of annotation
    annotation : dict
        overrides annotation_id if given
    groups_to_add : list
        each entry is a dict containing the information about user groups
        to add and their permission levels. A sample entry must have the
        following keys
        - level, int -> 0 (view), 1 (edit) or 2 (owner)
        - name, str -> name of user group
        - id, st -> girder id of user group
    replace_original_groups : bool
        whether to replace original groups or append to them
    users_to_add : list
        each entry is a dict containing the information about user
        to add and their permission levels. A sample entry must have the
        following keys
        - level, int -> 0 (view), 1 (edit) or 2 (owner)
        - login, str -> username of user
        - id, st -> girder id of user
    replace_original_users
        whether to replace original users or append to them

    Returns
    -------
    dict
        server response

    """
    groups_to_add = [] if groups_to_add is None else groups_to_add
    users_to_add = [] if users_to_add is None else users_to_add

    if annotation is not None:
        annotation_id = annotation['_id']
    elif annotation_id is None:
        msg = 'You must provide either the annotation or its girder id.'
        raise Exception(
            msg)

    # get current permissions
    current = gc.get('/annotation/%s/access' % annotation_id)

    # add or replace as needed
    if replace_original_groups:
        current['groups'] = []
        current_group_ids = []
    else:
        current_group_ids = [j['id'] for j in current['groups']]

    if replace_original_users:
        current['users'] = []
        current_user_ids = []
    else:
        current_user_ids = [j['id'] for j in current['users']]

    for group in groups_to_add:
        if group['id'] not in current_group_ids:
            current['groups'].append(group)

    for user in users_to_add:
        if user['id'] not in current_user_ids:
            current['users'].append(user)

    # now update accordingly
    # BAD WAY!! -- do NOT do this!
    # return gc.put('/annotation/%s/access?access=%s' % (
    #        annotation_id, json.dumps(current)))
    # PROPER WAY
    return gc.put('/annotation/%s/access' % annotation_id, data={
        'access': json.dumps(current)})


def update_permissions_for_annotations_in_slide(
        gc, slide_id, verbose=0, monitorPrefix='', **kwargs):
    """Update permissions for all annotations in a slide.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    slide_id : str
        girder id of slide
    verbose : int
        level of verbosity
    monitorPrefix : str
        prefix to prepend to printed statements
    kwargs
        passed as-is to update_permissions_for_annotation()

    Returns
    -------
    list
        each entry is a dict of the server response.

    """
    anniter = Annotation_iterator(
        gc=gc, slide_id=slide_id,
        callback=update_permissions_for_annotation,
        callback_kwargs=kwargs,
        verbose=verbose, monitorPrefix=monitorPrefix)
    return anniter.apply_callback_to_all_annotations()


def update_permissions_for_annotations_in_folder(
        gc, folderid, workflow_kwargs, recursive=True,
        monitor='', verbose=True):
    """Update permissions for all annotations in a folder recursively.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    folderid : str
        girder id of folder
    workflow_kwargs : dict
        kwargs to pass to update_permissions_for_annotations_in_slide()
    recursive : bool
        do this recursively for subfolders?
    monitor : str
        text to prepend to printed statements
    verbose : bool
        print statements to screen?

    Returns
    -------
    None

    """
    # update permissions for each slide in folder
    workflow_kwargs.update({'gc': gc})
    workflow_runner = Workflow_runner(
        slide_iterator=Slide_iterator(
            gc, source_folder_id=folderid,
            keep_slides=None,
        ),
        workflow=update_permissions_for_annotations_in_slide,
        workflow_kwargs=workflow_kwargs,
        recursive=recursive,
        monitorPrefix=monitor,
        verbose=verbose,
    )
    workflow_runner.run()


def update_styles_for_annotation(gc, annotation, changes):
    """Update styles for all relevant elements in an annotation.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    annotation : dict
        annotation
    changes : dict
        indexed by current group name to be updated, and values are
        the new styles. Each element in ann["annotation"]["elements"]
        whose current "group" attribute is in this dict's keys is
        updated according to the new style.

    Returns
    -------
    dict
        server response

    """
    # find out if annotation needs editing
    if 'groups' not in annotation.keys():
        return
    elif not any(g in changes.keys() for g in annotation['groups']):
        return

    # edit elements one by one
    for el in annotation['annotation']['elements']:
        if el['group'] in changes.keys():
            el.update(changes[el['group']])
    # print("  updating ...")
    return gc.put(
        '/annotation/%s' % annotation['_id'], json=annotation['annotation'])


def update_styles_for_annotations_in_slide(
        gc, slide_id, verbose=0, monitorPrefix='', callback=None, **kwargs):
    """Update styles for all annotations in a slide.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    slide_id : str
        girder id of slide
    verbose : int
        level of verbosity
    monitorPrefix : str
        prefix to prepend to printed statements
    callback : function
        if None, defaults to update_styles_for_annotation. Passed as-is
        to histomicstk.workflows.workflow_runner.Annotation_iterator
    kwargs
        passed as-is to the update_styles_for_annotation

    Returns
    -------
    list
        each entry is a dict of the server response.

    """
    if callback is None:
        callback = update_styles_for_annotation
    anniter = Annotation_iterator(
        gc=gc, slide_id=slide_id,
        callback=callback,
        callback_kwargs=kwargs,
        verbose=verbose, monitorPrefix=monitorPrefix)
    return anniter.apply_callback_to_all_annotations()


def update_styles_for_annotations_in_folder(
        gc, folderid, workflow_kwargs, recursive=True,
        catch_exceptions=True, monitor='', verbose=True):
    """Update styles for all annotations in a folder recursively.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    folderid : str
        girder id of folder
    workflow_kwargs : dict
        kwargs to pass to Update styles for all annotations in a slide()
    recursive : bool
        do this recursively for subfolders?
    catch_exceptions : bool
        passed as-is to histomicstk.workflows.workflow_runner.Workflow_runner
    monitor : str
        text to prepend to printed statements
    verbose : bool
        print statements to screen?

    Returns
    -------
    None

    """
    # update annotation styles
    workflow_kwargs.update({'gc': gc})
    workflow_runner = Workflow_runner(
        slide_iterator=Slide_iterator(
            gc, source_folder_id=folderid,
            keep_slides=None,
        ),
        workflow=update_styles_for_annotations_in_slide,
        workflow_kwargs=workflow_kwargs,
        recursive=recursive,
        catch_exceptions=catch_exceptions,
        monitorPrefix=monitor,
        verbose=verbose,
    )
    workflow_runner.run()


def revert_annotation(
        gc, annotation_id=None, annotation=None, version=None,
        revert_to_nonempty_elements=False, only_revert_if_empty=True):
    """Revert an annotation to a previous version.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    annotation_id : str
        girder id of annotation
    annotation : dict
        overrides annotation_id if given
    version : int
        version number for annotation. If None, and
        not revert_to_nonempty_elements
        the default behavior of the endpoint is evoked, which reverts the
        annotation if it was deleted and if not, reverts to the last version.
    revert_to_nonempty_elements : bool
        if true, reverts to the most recent version of the annotation
        with non-empty elements.
    only_revert_if_empty : bool
        if true, only reverts annotation if it contains an empty element list

    Returns
    -------
    dict
        server response

    """
    if annotation is not None:
        annotation_id = annotation['_id']
    elif annotation_id is None:
        msg = 'You must provide either the annotation or its girder id.'
        raise Exception(
            msg)

    history = gc.get('/annotation/%s/history' % annotation_id)

    # no need to revert if empty
    if only_revert_if_empty and len(history[0]['groups']) > 0:
        return {}

    # cannot revert if only version
    if len(history) < 2:
        return {}

    if (version is None) and revert_to_nonempty_elements:

        # NOTE: even though the "history" may show
        # the elements as empty, the "groups" attribute is really the
        # indication if the annotation version actually has some elements.
        # TODO -- This is likely a bug (?); fix me!!!
        for ver in history:
            if len(ver['groups']) > 0:
                version = ver['_version']
                break

    ver = '' if version is None else '?version=%d' % version

    # if version is None:
    #     print("    Reverting ...")
    # else:
    #     print("    Reverting to version %d" % version)

    return gc.put('/annotation/%s/history/revert%s' % (annotation_id, ver))


def revert_annotations_in_slide(
        gc, slide_id, verbose=0, monitorPrefix='', **kwargs):
    """Revert all annotations in a slide to a previous version.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    slide_id : str
        girder id of slide
    verbose : int
        level of verbosity
    monitorPrefix : str
        prefix to prepend to printed statements
    kwargs
        passed as-is to the revert_annotation

    Returns
    -------
    list
        each entry is a dict of the server response.

    """
    anniter = Annotation_iterator(
        gc=gc, slide_id=slide_id,
        callback=revert_annotation,
        callback_kwargs=kwargs,
        verbose=verbose, monitorPrefix=monitorPrefix)
    return anniter.apply_callback_to_all_annotations()


def revert_annotations_in_folder(
        gc, folderid, workflow_kwargs, recursive=True,
        monitor='', verbose=True):
    """Revert all annotations in a folder recursively.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client
    folderid : str
        girder id of folder
    workflow_kwargs : dict
        kwargs to pass to revert_annotations_in_slide
    recursive : bool
        do this recursively for subfolders?
    monitor : str
        text to prepend to printed statements
    verbose : bool
        print statements to screen?

    Returns
    -------
    None

    """
    # update annotation styles
    workflow_kwargs.update({'gc': gc})
    workflow_runner = Workflow_runner(
        slide_iterator=Slide_iterator(
            gc, source_folder_id=folderid,
            keep_slides=None,
        ),
        workflow=revert_annotations_in_slide,
        workflow_kwargs=workflow_kwargs,
        recursive=recursive,
        monitorPrefix=monitor,
        verbose=verbose,
    )
    workflow_runner.run()


def reproduce_annotations_workflow(
        gc, folderid, annotation_jsonfile, monitorPrefix=''):
    """Dump annotations into single slide from local folder (Internal function).

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client instance

    folderid : str
        girder id of girder folder to copy the slide to before posting
        the annotations to it. This prevents confusing old annotations that
        were associated with the slide originally and the new posted annotations.

    annotation_jsonfile : str
        path to annotation json file

    monitorPrefix : str
        prefix to monitor string

    """
    try:
        # extract name + path
        itemname = os.path.basename(annotation_jsonfile).replace(
            '_annotations.json', '')
        local = os.path.dirname(annotation_jsonfile)

        # copy item without annotations
        with open(os.path.join(local, itemname + '.json')) as jf:
            source_item_info = json.load(jf)
        print('%s: copy item' % monitorPrefix)
        item = gc.post(
            '/item/%s/copy?folderId=%s&name=%s&copyAnnotations=False'
            % (source_item_info['_id'], folderid, itemname))

        # load annotations
        with open(annotation_jsonfile) as af:
            annotations = json.load(af)

        # now post
        n_annotations = len(annotations)
        for anno, annotation in enumerate(annotations):
            try:
                print('%s: post annotation %d of %d' % (
                    monitorPrefix, anno, n_annotations))
                _ = gc.post(
                    '/annotation?itemId=' + item['_id'],
                    json=annotation['annotation'])
            except Exception as e:
                print(e.__repr__())

    except Exception as e:
        print(e.__repr__())


def reproduce_annotations_from_backup(gc, folderid, local):
    """Reproduce annotations on HistomicsUI from local backup.

    This is the reverse functionality to dump_annotations.locally().
    This reproduces this tiered structure on HistomicsUI. The original
    slides (items) must still be there in the folder from the backup was
    made because these will be copied (without annotations) before the
    local annotations (from JSON files) are posted to them.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client instance

    folderid : str
        girder id of girder folder to post reproduced annotations.

    local : str
        local path to get subfolders/slides/annotations

    """
    monitor = os.path.basename(local)

    # for each slide, copy it and post annotations
    jsonfiles = [
        os.path.join(local, j) for j in os.listdir(local)
        if j.endswith('_annotations.json')]
    for jsonfile in jsonfiles:
        reproduce_annotations_workflow(
            gc=gc, folderid=folderid, annotation_jsonfile=jsonfile,
            monitorPrefix=monitor)

    # for each subfolder, create a new folder on HistomicsUI and call self
    subdirs = [
        j for j in os.listdir(local) if os.path.isdir(os.path.join(local, j))]
    for subdir in subdirs:
        try:
            # create folder in HistomicsUI
            new_folder = gc.post('/folder?parentId=%s&name=%s' % (
                folderid, subdir))

            # call self with same parameters
            reproduce_annotations_from_backup(
                gc=gc, folderid=new_folder['_id'],
                local=os.path.join(local, subdir))

        except Exception as e:
            print(e.__repr__())



"""Compute statistics for reinhard normalization."""

import numpy as np

from .rgb_to_lab import rgb_to_lab


def lab_mean_std(im_input, mask_out=None):
    """Compute the mean and standard deviation of the intensities.

    ... of each channel of the given RGB image in LAB color space.
    The outputs of this function is for reinhard normalization.

    Parameters
    ----------
    im_input : array_like
        An RGB image

    mask_out : array_like
        if not None, uses numpy masked array functionality to only keep
        non-masked areas when calculating mean and standard deviation.

    Returns
    -------
    mean_lab : array_like
        A 3-element array containing the mean of each channel of the input RGB
        in LAB color space.

    std_lab : array_like
        A 3-element array containing the standard deviation of each channel
        of the input RGB in LAB color space.

    mask_out : array_like, default is None
        if not None, should be (m, n) boolean numpy array.
        This method uses numpy masked array functionality to only use
        non-masked areas in calculations. This is relevant because elements
        like blood, sharpie marker, white space, etc would throw off the
        reinhard normalization by affecting the mean and stdev. Ideally, you
        want to exclude these elements from both the target image (from which
        you calculate target_mu and target_sigma) and from the source image
        to be normalized.

    See Also
    --------
    histomicstk.preprocessing.color_conversion.rgb_to_lab,
    histomicstk.preprocessing.color_conversion.reinhard

    References
    ----------
    .. [#] E. Reinhard, M. Adhikhmin, B. Gooch, P. Shirley, "Color transfer
       between images," in IEEE Computer Graphics and Applications, vol.21,
       no.5,pp.34-41, 2001.
    .. [#] D. Ruderman, T. Cronin, and C. Chiao, "Statistics of cone
       responses to natural images: implications for visual coding,"
       J. Opt. Soc. Am. A vol.15, pp.2036-2045, 1998.

    """
    im_lab = rgb_to_lab(im_input)

    # mask out irrelevant tissue / whitespace / etc
    if mask_out is not None:
        mask_out = mask_out[..., None]
        im_lab = np.ma.masked_array(
            im_lab, mask=np.tile(mask_out, (1, 1, 3)))

    mean_lab = np.array([im_lab[..., i].mean() for i in range(3)])
    std_lab = np.array([im_lab[..., i].std() for i in range(3)])

    return mean_lab, std_lab



from argparse import Namespace
from datetime import timedelta

import large_image
import numpy as np
from ctk_cli import CLIArgumentParser  # noqa I004
# imported for side effects
from slicer_cli_web import ctk_cli_adjustment  # noqa

import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.segmentation as htk_seg
import histomicstk.utils as htk_utils

# These defaults are only used if girder is not present
# Use memcached by default.
large_image.config.setConfig('cache_backend', 'memcached')
# If memcached is unavailable, specify the fraction of memory that python
# caching is allowed to use.  This is deliberately small.
large_image.config.setConfig('cache_python_memory_portion', 32)


def get_stain_vector(args, index):
    """Get the stain corresponding to args.stain_$index and
    args.stain_$index_vector.  If the former is not "custom", all the
    latter's elements must be -1.

    """
    args = args._asdict() if hasattr(args, '_asdict') else vars(args)
    stain = args['stain_' + str(index)]
    stain_vector = args['stain_' + str(index) + '_vector']
    if all(x == -1 for x in stain_vector):  # Magic default value
        if stain == 'custom':
            msg = (
                'If "custom" is chosen for a stain, '
                'a stain vector must be provided.'
            )
            raise ValueError(msg)
        return htk_cdeconv.stain_color_map[stain]
    else:
        if stain == 'custom':
            return stain_vector
        msg = (
            'Unless "custom" is chosen for a stain, '
            'no stain vector may be provided.'
        )
        raise ValueError(msg)


def get_stain_matrix(args, count=3):
    """Get the stain matrix corresponding to the args.stain_$index and
    args.stain_$index_vector arguments for values of index 1 to count.
    Return a numpy array of column vectors.

    """
    return np.array([get_stain_vector(args, i + 1) for i in range(count)]).T


def segment_wsi_foreground_at_low_res(
        ts, lres_size=2048, invert_image=False, frame=None, default_img_inversion=False):

    ts_metadata = ts.getMetadata()

    # get image at low-res
    maxSize = max(ts_metadata['sizeX'], ts_metadata['sizeY'])
    maxSize = float(max(maxSize, lres_size))

    downsample_factor = 2.0 ** np.floor(np.log2(maxSize / lres_size))

    fgnd_seg_mag = ts_metadata['magnification'] / downsample_factor

    fgnd_seg_scale = {'magnification': fgnd_seg_mag}

    im_lres, _ = ts.getRegion(
        scale=fgnd_seg_scale,
        format=large_image.tilesource.TILE_FORMAT_NUMPY,
        frame=frame,
    )

    # check number of channels
    if len(im_lres.shape) <= 2 or im_lres.shape[2] == 1:
        im_lres = np.dstack((im_lres, im_lres, im_lres))
        if default_img_inversion:
            invert_image = True
    else:
        im_lres = im_lres[:, :, :3]

    # perform image inversion
    if invert_image:
        im_lres = np.max(im_lres) - im_lres

    # casting the float to integers
    if issubclass(im_lres.dtype.type, np.floating) and np.max(im_lres) > 1:
        if np.min(im_lres) >= 0 and np.max(im_lres) < 256:
            im_lres = im_lres.astype(np.uint8)
        elif np.min(im_lres) >= 0 and np.max(im_lres) < 65536:
            im_lres = im_lres.astype(np.uint16)

    # compute foreground mask at low-res
    im_fgnd_mask_lres = htk_utils.simple_mask(im_lres)

    return im_fgnd_mask_lres, fgnd_seg_scale


def create_tile_nuclei_bbox_annotations(im_nuclei_seg_mask, tile_info, region_props=None):
    import skimage.measure

    nuclei_annot_list = []

    gx = tile_info['gx']
    gy = tile_info['gy']
    wfrac = tile_info['gwidth'] / np.double(tile_info['width'])
    hfrac = tile_info['gheight'] / np.double(tile_info['height'])

    nuclei_obj_props = region_props if region_props else skimage.measure.regionprops(
        im_nuclei_seg_mask)

    for i in range(len(nuclei_obj_props)):
        cx = nuclei_obj_props[i].centroid[1]
        cy = nuclei_obj_props[i].centroid[0]
        width = nuclei_obj_props[i].bbox[3] - nuclei_obj_props[i].bbox[1] + 1
        height = nuclei_obj_props[i].bbox[2] - nuclei_obj_props[i].bbox[0] + 1

        # convert to base pixel coords
        cx = np.round(gx + cx * wfrac, 2)
        cy = np.round(gy + cy * hfrac, 2)
        width = np.round(width * wfrac, 2)
        height = np.round(height * hfrac, 2)

        # create annotation json
        cur_bbox = {
            'type': 'rectangle',
            'center': [cx, cy, 0],
            'width': width,
            'height': height,
            'rotation': 0,
            'fillColor': 'rgba(0,0,0,0)',
            'lineColor': 'rgb(0,255,0)',
        }

        nuclei_annot_list.append(cur_bbox)

    return nuclei_annot_list


def create_tile_nuclei_boundary_annotations(im_nuclei_seg_mask, tile_info, region_props=None):

    nuclei_annot_list = []

    gx = tile_info['gx']
    gy = tile_info['gy']
    wfrac = tile_info['gwidth'] / np.double(tile_info['width'])
    hfrac = tile_info['gheight'] / np.double(tile_info['height'])

    if region_props:
        by, bx, selected_rows = htk_seg.label.trace_object_boundaries(im_nuclei_seg_mask,
                                                                      trace_all=True,
                                                                      region_props=region_props)
    else:
        by, bx = htk_seg.label.trace_object_boundaries(im_nuclei_seg_mask,
                                                       trace_all=True)
    for i in range(len(bx)):

        # get boundary points and convert to base pixel space
        num_points = len(bx[i])

        if num_points < 3:
            continue

        cur_points = np.zeros((num_points, 3))
        cur_points[:, 0] = np.round(gx + bx[i] * wfrac, 2)
        cur_points[:, 1] = np.round(gy + by[i] * hfrac, 2)
        cur_points = cur_points.tolist()

        # create annotation json
        cur_annot = {
            'type': 'polyline',
            'points': cur_points,
            'closed': True,
            'fillColor': 'rgba(0,0,0,0)',
            'lineColor': 'rgb(0,255,0)',
        }

        nuclei_annot_list.append(cur_annot)
    if region_props:
        return nuclei_annot_list, selected_rows
    return nuclei_annot_list


def create_tile_nuclei_annotations(im_nuclei_seg_mask, tile_info, format, region_props=None):

    if format == 'bbox':

        return create_tile_nuclei_bbox_annotations(im_nuclei_seg_mask,
                                                   tile_info, region_props)

    elif format == 'boundary':

        return create_tile_nuclei_boundary_annotations(im_nuclei_seg_mask,
                                                       tile_info, region_props)
    else:

        msg = 'Invalid value passed for nuclei_annotation_format'
        raise ValueError(msg)


def create_dask_client(args):
    """Create and install a Dask distributed client using args from a
    Namespace, supporting the following attributes:

    - .scheduler: Address of the distributed scheduler, or the
      empty string to start one locally

    """
    # Dask is noisy.  Stop that
    import os
    os.environ.setdefault('DASK_DISTRIBUTED__LOGGING__DISTRIBUTED', 'warning')

    import dask
    import psutil

    scheduler = getattr(args, 'scheduler', None)
    num_workers = getattr(args, 'num_workers', 0)
    num_threads_per_worker = getattr(args, 'num_threads_per_worker', 0)

    if scheduler == 'multithreading':
        from multiprocessing.pool import ThreadPool

        import dask.threaded

        if num_threads_per_worker <= 0:
            num_workers = max(1, psutil.cpu_count(logical=False) + num_threads_per_worker)
        else:
            num_workers = num_threads_per_worker
        print('Starting dask thread pool with %d thread(s)' % num_workers)
        dask.config.set(pool=ThreadPool(num_workers))
        dask.config.set(scheduler='threads')
        return

    if scheduler == 'multiprocessing':
        import multiprocessing

        import dask.multiprocessing

        dask.config.set(scheduler='processes')
        if num_workers <= 0:
            num_workers = max(1, psutil.cpu_count(logical=False) + num_workers)

        print('Starting dask multiprocessing pool with %d worker(s)' % num_workers)
        dask.config.set(pool=multiprocessing.Pool(
            num_workers, initializer=dask.multiprocessing.initialize_worker_process))
        return

    import dask.distributed
    if not scheduler:

        if num_workers <= 0:
            num_workers = max(1, psutil.cpu_count(logical=False) + num_workers)
        num_threads_per_worker = (num_threads_per_worker if num_threads_per_worker >= 1 else None)

        print('Creating dask LocalCluster with %d worker(s), %r thread(s) per '
              'worker' % (num_workers, num_threads_per_worker))
        scheduler = dask.distributed.LocalCluster(
            ip='0.0.0.0',  # Allow reaching the diagnostics port externally
            scheduler_port=0,  # Don't expose the scheduler port
            n_workers=num_workers,
            memory_limit=0,
            threads_per_worker=num_threads_per_worker,
            silence_logs=False,
        )

    return dask.distributed.Client(scheduler)


def get_region_polygons(region):
    """
    Convert the region into a list of polygons.

    Params
    ------
    region: list
        4 elements -- left, top, width, height -- or all -1, meaning the whole
        slide.
        6 elements -- x,y,z,rx,ry,rz: a rectangle specified by center and
        radius.
        8 or more elements -- x,y,x,y,...: a polygon.  The region is the
        bounding box.

    Returns
    -------
    polygons: list
        A list of lists of x, y tuples.

    """
    if len(region) % 2 or len(region) < 4:
        msg = 'region must be 4, 6, or a list of 2n values.'
        raise ValueError(msg)
    region = [float(v) for v in region]
    if region == [-1] * 4:
        return []
    if len(region) == 6:
        region = [region[0] - region[3], region[1] - region[4], region[3] * 2, region[4] * 2]
    if len(region) == 4:
        region = [
            region[0], region[1],
            region[0] + region[2], region[1],
            region[0] + region[2], region[1] + region[3],
            region[0], region[1] + region[3],
        ]
    polys = [[]]
    for idx, x in enumerate(region[::2]):
        y = region[idx * 2 + 1]
        if x == -1 and y == -1:
            polys.append([])
        elif not len(polys[-1]) or (x, y) != tuple(polys[-1][-1]):
            polys[-1].append((x, y))
    for poly in polys:
        if len(poly) and poly[0] == poly[-1]:
            poly[-1:] = []
    polys = [poly for poly in polys if len(poly) >= 3]
    return polys


def polygons_to_binary_mask(polygons, x=0, y=0, width=None, height=None):
    """Convert a set of region polygons to a numpy binary mask.

    Params
    ------
    polygons: list
        A list of lists of x, y tuples.  If None, returns None.
    x: integer
        An offset for the mask compared to the polygon coordinates.
    y: integer
        An offset for the mask compared to the polygon coordinates.
    width: integer
        The width of the mask to return.  None uses the maximum polygon
        coordinate.
    height: integer
        The height of the mask to return.  None uses the maximum polygon
        coordinate.

    Returns
    -------
    mask: numpy.array
        A 1-bit numpy array where 1 is inside an odd number of polygons.  This
        can return None if polygons was None.

    """
    import PIL.Image
    import PIL.ImageChops
    import PIL.ImageDraw

    if polygons is None or not len(polygons):
        return None
    if width is None:
        width = max(pt[0] for poly in polygons for pt in poly) - x
    if height is None:
        height = max(pt[1] for poly in polygons for pt in poly) - y
    full = PIL.Image.new('1', (width, height), 0)
    for polyidx, poly in enumerate(polygons):
        mask = PIL.Image.new('1', (width, height), 0)
        PIL.ImageDraw.Draw(mask).polygon(
            [(int(pt[0] - x), int(pt[1] - y)) for pt in poly], outline=None, fill=1)
        if not polyidx:
            full = mask
        else:
            full = PIL.ImageChops.logical_xor(full, mask)
    return np.array(full)


def get_region_dict(region, maxRegionSize=None, tilesource=None):
    """Return a dict corresponding to region, checking the region size if
    maxRegionSize is provided.

    The intended use is to be passed via **kwargs, and so either {} is
    returned (for the special region -1,-1,-1,-1) or {'region':
    region_dict}.

    Params
    ------
    region: list
        4 elements -- left, top, width, height -- or all -1, meaning the whole
        slide.
        6 elements -- x,y,z,rx,ry,rz: a rectangle specified by center and
        radius.
        8 or more elements -- x,y,x,y,...: a polygon.  The region is the
        bounding box.
    maxRegionSize: int, optional
        Maximum size permitted of any single dimension.  -1 or None for
        unlimited.
    tilesource: tilesource, optional
        A `large_image` tilesource (or anything with `.sizeX` and `.sizeY`
        properties) that is used to determine the size of the whole slide if
        necessary.  Must be provided if `maxRegionSize` is not None.

    Returns
    -------
    region_dict: dict
        Either {} (for the special region -1,-1,-1,-1) or
        {'region': region_subdict}

    """
    if len(region) % 2 or len(region) < 4:
        msg = 'region must be 4, 6, or a list of 2n values.'
        raise ValueError(msg)

    useWholeImage = region == [-1] * 4

    if len(region) == 6:
        region = [region[0] - region[3], region[1] - region[4], region[3] * 2, region[4] * 2]
    if len(region) >= 8:
        polygons = get_region_polygons(region)
        minx = min(pt[0] for poly in polygons for pt in poly)
        maxx = max(pt[0] for poly in polygons for pt in poly)
        miny = min(pt[1] for poly in polygons for pt in poly)
        maxy = max(pt[1] for poly in polygons for pt in poly)
        region = [minx, miny, maxx - minx, maxy - miny]

    if maxRegionSize is not None and maxRegionSize > 0:
        if tilesource is None:
            msg = 'tilesource must be provided if maxRegionSize is specified'
            raise ValueError(msg)
        if useWholeImage:
            size = max(tilesource.sizeX, tilesource.sizeY)
        else:
            size = max(region[-2:])
        if size > maxRegionSize:
            msg = (
                'Requested region is too large!  '
                'Please see --maxRegionSize'
            )
            raise ValueError(msg)
    # If a tilesource was specified, restrict the region to the image size
    if not useWholeImage and tilesource:
        minx = max(0, region[0])
        maxx = min(tilesource.sizeX, region[0] + region[2])
        miny = max(0, region[1])
        maxy = min(tilesource.sizeY, region[1] + region[3])
        region = [minx, miny, maxx - minx, maxy - miny]

    return {} if useWholeImage else dict(
        region=dict(zip(['left', 'top', 'width', 'height'],
                        [int(val) for val in region])))


def disp_time_hms(seconds):
    """Converts time from seconds to a string of the form hours:minutes:seconds
    """
    return str(timedelta(seconds=seconds))


def splitArgs(args, split='_'):
    """Split a Namespace into a Namespace of Namespaces based on shared
    prefixes.  The string separating the prefix from the rest of the
    argument is determined by the optional "split" parameter.
    Parameters not containing the splitting string are kept as-is.

    """
    def splitKey(k):
        s = k.split(split, 1)
        return (None, s[0]) if len(s) == 1 else s

    args = args._asdict() if hasattr(args, '_asdict') else vars(args)
    firstKeys = {splitKey(k)[0] for k in args}
    result = Namespace()
    for k in firstKeys - {None}:
        setattr(result, k, Namespace())
    for k, v in args.items():
        f, s = splitKey(k)
        if f is None:
            setattr(result, s, v)
        else:
            setattr(getattr(result, f), s, v)
    return result


def sample_pixels(args):
    """Version of histomicstk.utils.sample_pixels that takes a Namespace
    and handles the special default values.

    """
    args = (args._asdict() if hasattr(args, '_asdict') else vars(args)).copy()
    for k in 'magnification', 'sample_fraction', 'sample_approximate_total':
        if args[k] == -1:
            del args[k]
    return htk_utils.sample_pixels(**args)


__all__ = (
    'CLIArgumentParser',
    'create_dask_client',
    'create_tile_nuclei_annotations',
    'create_tile_nuclei_bbox_annotations',
    'create_tile_nuclei_boundary_annotations',
    'disp_time_hms',
    'get_region_dict',
    'get_stain_matrix',
    'get_stain_vector',
    'sample_pixels',
    'segment_wsi_foreground_at_low_res',
    'splitArgs',
)



import numpy as np

from .condense import condense
from .delete import delete


def delete_overlap(im_label, overlap_info):
    """
    Deletes overlapping regions from an image label based on overlap information and tile size.

    Args:
    ----
        im_label (ndarray): Image label represented as a NumPy array.
        overlap_info (dict): Dictionary containing overlap information.
                            It should have the following keys:
                            'left', 'right', 'top', and 'bottom',
                            each specifying the overlap amount in pixels.

    Returns:
    -------
        ndarray: Image label with overlapping regions deleted.

    Note:
    ----
        This function assumes the necessary imports, such as `np`, are already present.

    """
    # Compute the half of the overlap values
    left_overlap = overlap_info['left'] // 2
    right_overlap = overlap_info['right'] // 2
    top_overlap = overlap_info['top'] // 2
    bottom_overlap = overlap_info['bottom'] // 2

    if not np.any(im_label):
        return im_label

    # Create a border mask based on the overlap values
    im_border_mask = np.zeros_like(im_label)
    if left_overlap:
        im_border_mask[:, :left_overlap] = True
    if right_overlap:
        im_border_mask[:, -right_overlap:] = True
    if top_overlap:
        im_border_mask[:top_overlap, :] = True
    if bottom_overlap:
        im_border_mask[-bottom_overlap:, :] = True

    # Find unique indices of the border regions
    border_indices = np.unique(im_label[im_border_mask > 0])
    border_indices = border_indices[border_indices > 0]

    if len(border_indices) == 0:
        return im_label

    # Condense and delete the overlap regions from the image label
    im_label_del = np.zeros_like(im_label)
    im_label_del = condense(delete(im_label, border_indices))
    return im_label_del



"""
# Modified from: https://code.google.com/archive/p/pyrtree/.

# Copyright Google
# Under The 3-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION). HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""

import array
import random
import time

from .rect import NullRect, Rect, union_all

MAXCHILDREN = 10
MAX_KMEANS = 5


class RTree:
    """Placeholder."""

    def __init__(self):
        """Placeholder."""
        self.count = 0
        self.stats = {
            'overflow_f': 0,
            'avg_overflow_t_f': 0.0,
            'longest_overflow': 0.0,
            'longest_kmeans': 0.0,
            'sum_kmeans_iter_f': 0,
            'count_kmeans_iter_f': 0,
            'avg_kmeans_iter_f': 0.0,
        }

        # This round: not using objects directly -- they
        #   take up too much memory, and efficiency goes down the toilet
        #   (obviously) if things start to page.
        #  Less obviously: using object graph directly leads to really long GC
        #   pause times, too.
        # Instead, it uses pools of arrays:
        self.count = 0
        self.leaf_count = 0
        self.rect_pool = array.array('d')
        self.node_pool = array.array('L')
        self.leaf_pool = []  # leaf objects.

        self.cursor = _NodeCursor.create(self, NullRect)

    def _ensure_pool(self, idx):
        if len(self.rect_pool) < (4 * idx):
            self.rect_pool.extend([0, 0, 0, 0] * idx)
            self.node_pool.extend([0, 0] * idx)

    def insert(self, o, orect):
        """Placeholder."""
        self.cursor.insert(o, orect)
        assert self.cursor.index == 0

    def query_rect(self, r):
        """Placeholder."""
        yield from self.cursor.query_rect(r)

    def query_point(self, p):
        """Placeholder."""
        yield from self.cursor.query_point(p)

    def walk(self, pred):
        """Placeholder."""
        return self.cursor.walk(pred)


class _NodeCursor:
    """Placeholder."""

    @classmethod
    def create(cls, rooto, rect):
        """Placeholder."""
        idx = rooto.count
        rooto.count += 1

        rooto._ensure_pool(idx + 1)
        # rooto.node_pool.extend([0,0])
        # rooto.rect_pool.extend([0,0,0,0])

        retv = _NodeCursor(rooto, idx, rect, 0, 0)

        retv._save_back()
        return retv

    @classmethod
    def create_with_children(cls, children, rooto):
        """Placeholder."""
        rect = union_all(list(children))
        # nr = Rect(rect.x,rect.y,rect.xx,rect.yy)
        assert not rect.swapped_x
        nc = _NodeCursor.create(rooto, rect)
        nc._set_children(children)
        assert not nc.is_leaf()
        return nc

    @classmethod
    def create_leaf(cls, rooto, leaf_obj, leaf_rect):
        """Placeholder."""
        rect = Rect(leaf_rect.x, leaf_rect.y, leaf_rect.xx, leaf_rect.yy)
        rect.swapped_x = True  # Mark as leaf by setting the xswap flag.
        res = _NodeCursor.create(rooto, rect)
        idx = res.index
        res.first_child = rooto.leaf_count
        rooto.leaf_count += 1
        res.next_sibling = 0
        rooto.leaf_pool.append(leaf_obj)
        res._save_back()
        res._become(idx)
        assert res.is_leaf()
        return res

    __slots__ = (
        'root', 'npool', 'rpool', 'index', 'rect',
        'next_sibling', 'first_child')

    def __init__(self, rooto, index, rect, first_child, next_sibling):
        """Placeholder."""
        self.root = rooto
        self.rpool = rooto.rect_pool
        self.npool = rooto.node_pool

        self.index = index
        self.rect = rect
        self.next_sibling = next_sibling
        self.first_child = first_child

    def walk(self, predicate):
        """Placeholder."""
        if (predicate(self, self.leaf_obj())):
            yield self
            if not self.is_leaf():
                for c in self.children():
                    yield from c.walk(predicate)

    def query_rect(self, r):
        """Return things that intersect with 'r'."""
        def p(o, x):
            return r.does_intersect(o.rect)
        yield from self.walk(p)

    def query_point(self, point):
        """Query by a point."""
        def p(o, x):
            return o.rect.does_containpoint(point)

        yield from self.walk(p)

    def lift(self):
        """Placeholder."""
        return _NodeCursor(self.root,
                           self.index,
                           self.rect,
                           self.first_child,
                           self.next_sibling)

    def _become(self, index):
        """Placeholder."""
        recti = index * 4
        nodei = index * 2
        rp = self.rpool
        x = rp[recti]
        y = rp[recti + 1]
        xx = rp[recti + 2]
        yy = rp[recti + 3]

        if (x == 0.0 and y == 0.0 and xx == 0.0 and yy == 0.0):
            self.rect = NullRect
        else:
            self.rect = Rect(x, y, xx, yy)

        self.next_sibling = self.npool[nodei]
        self.first_child = self.npool[nodei + 1]
        self.index = index

    def is_leaf(self):
        """Placeholder."""
        return self.rect.swapped_x

    def has_children(self):
        """Placeholder."""
        return (not self.is_leaf()) and (0 != self.first_child)

    def holds_leaves(self):
        """Placeholder."""
        if 0 == self.first_child:
            return True
        else:
            return self.has_children() and self.get_first_child().is_leaf()

    def get_first_child(self):
        """Placeholder."""
        # fc = self.first_child
        c = _NodeCursor(self.root, 0, NullRect, 0, 0)
        c._become(self.first_child)
        return c

    def leaf_obj(self):
        """Placeholder."""
        if self.is_leaf():
            return self.root.leaf_pool[self.first_child]
        else:
            return None

    def _save_back(self):
        """Placeholder."""
        rp = self.rpool
        recti = self.index * 4
        nodei = self.index * 2

        if self.rect is not NullRect:
            self.rect.write_raw_coords(rp, recti)
        else:
            rp[recti] = 0
            rp[recti + 1] = 0
            rp[recti + 2] = 0
            rp[recti + 3] = 0

        self.npool[nodei] = self.next_sibling
        self.npool[nodei + 1] = self.first_child

    def nchildren(self):
        """Placeholder."""
        # i = self.index
        c = 0
        for _x in self.children():
            c += 1
        return c

    def insert(self, leafo, leafrect):
        """Placeholder."""
        index = self.index

        # tail recursion, made into loop:
        while True:
            if self.holds_leaves():
                self.rect = self.rect.union(leafrect)
                self._insert_child(_NodeCursor.create_leaf(
                    self.root, leafo, leafrect))

                self._balance()

                # done: become the original again
                self._become(index)
                return
            else:
                # Not holding leaves, move down a level in the tree:

                # Micro-optimization:
                #  inlining union() calls -- logic is:
                # ignored,child = min(
                #    [ ((c.rect.union(leafrect)).area() - c.rect.area(),
                #    c.index) for c in self.children() ])
                child = None
                minarea = -1.0
                for c in self.children():
                    x, y, xx, yy = c.rect.coords()
                    lx, ly, lxx, lyy = leafrect.coords()
                    nx = x if x < lx else lx
                    nxx = xx if xx > lxx else lxx
                    ny = y if y < ly else ly
                    nyy = yy if yy > lyy else lyy
                    a = (nxx - nx) * (nyy - ny)
                    if minarea < 0 or a < minarea:
                        minarea = a
                        child = c.index
                # End micro-optimization

                self.rect = self.rect.union(leafrect)
                self._save_back()
                self._become(child)  # recurse.

    def _balance(self):
        """Placeholder."""
        if (self.nchildren() <= MAXCHILDREN):
            return

        t = time.time()

        s_children = [c.lift() for c in self.children()]

        memo = {}

        clusterings = [
            k_means_cluster(self.root, k, s_children)
            for k in range(2, MAX_KMEANS)]
        score, bestcluster = max([(
            silhouette_coeff(c, memo), c) for c in clusterings])

        nodes = [
            _NodeCursor.create_with_children(c, self.root)
            for c in bestcluster if len(c) > 0]

        self._set_children(nodes)

        dur = (time.time() - t)
        c = float(self.root.stats['overflow_f'])
        oa = self.root.stats['avg_overflow_t_f']
        self.root.stats['avg_overflow_t_f'] = (
            dur / (c + 1.0)) + (c * oa / (c + 1.0))
        self.root.stats['overflow_f'] += 1
        self.root.stats['longest_overflow'] = max(
            self.root.stats['longest_overflow'], dur)

    def _set_children(self, cs):
        """Placeholder."""
        self.first_child = 0

        if 0 == len(cs):
            return

        pred = None
        for c in cs:
            if pred is not None:
                pred.next_sibling = c.index
                pred._save_back()
            if 0 == self.first_child:
                self.first_child = c.index
            pred = c
        pred.next_sibling = 0
        pred._save_back()
        self._save_back()

    def _insert_child(self, c):
        """Placeholder."""
        c.next_sibling = self.first_child
        self.first_child = c.index
        c._save_back()
        self._save_back()

    def children(self):
        """Placeholder."""
        if (0 == self.first_child):
            return

        idx = self.index
        fc = self.first_child
        ns = self.next_sibling
        r = self.rect

        self._become(self.first_child)
        while True:
            yield self
            if 0 == self.next_sibling:
                break
            else:
                self._become(self.next_sibling)

        # Go back to becoming the same node we were.
        # self._become(idx)
        self.index = idx
        self.first_child = fc
        self.next_sibling = ns
        self.rect = r


def avg_diagonals(node, onodes, memo_tab):
    """Placeholder."""
    nidx = node.index
    sv = 0.0
    diag = 0.0
    for onode in onodes:
        k1 = (nidx, onode.index)
        k2 = (onode.index, nidx)
        if k1 in memo_tab:
            diag = memo_tab[k1]
        elif k2 in memo_tab:
            diag = memo_tab[k2]
        else:
            diag = node.rect.union(onode.rect).diagonal()
            memo_tab[k1] = diag

        sv += diag

    return sv / len(onodes)


def silhouette_w(node, cluster, next_closest_cluster, memo):
    """Placeholder."""
    ndist = avg_diagonals(node, cluster, memo)
    sdist = avg_diagonals(node, next_closest_cluster, memo)
    return (sdist - ndist) / max(sdist, ndist)


def silhouette_coeff(clustering, memo_tab):
    """Placeholder."""
    # special case for a clustering of 1.0
    if (len(clustering) == 1):
        return 1.0

    coeffs = []
    for cluster in clustering:
        others = [c for c in clustering if c is not cluster]
        others_cntr = [center_of_gravity(c) for c in others]
        ws = [silhouette_w(
            node, cluster, others[closest(others_cntr, node)], memo_tab)
            for node in cluster]
        cluster_coeff = sum(ws) / len(ws)
        coeffs.append(cluster_coeff)
    return sum(coeffs) / len(coeffs)


def center_of_gravity(nodes):
    """Placeholder."""
    totarea = 0.0
    xs, ys = 0, 0
    for n in nodes:
        if n.rect is not NullRect:
            x, y, w, h = n.rect.extent()
            a = w * h
            xs = xs + (a * (x + (0.5 * w)))
            ys = ys + (a * (y + (0.5 * h)))
            totarea = totarea + a
    return (xs / totarea), (ys / totarea)


def closest(centroids, node):
    """Placeholder."""
    x, y = center_of_gravity([node])
    dist = -1
    ridx = -1

    for (i, (xx, yy)) in enumerate(centroids):
        dsq = ((xx - x) ** 2) + ((yy - y) ** 2)
        if -1 == dist or dsq < dist:
            dist = dsq
            ridx = i
    return ridx


def k_means_cluster(root, k, nodes):
    """Placeholder."""
    t = time.time()
    if len(nodes) <= k:
        return [[n] for n in nodes]

    ns = list(nodes)
    root.stats['count_kmeans_iter_f'] += 1

    # Initialize: take n random nodes.
    random.shuffle(ns)

    cluster_centers = [center_of_gravity([n]) for n in ns[:k]]

    # Loop until stable:
    while True:
        root.stats['sum_kmeans_iter_f'] += 1
        clusters = [[] for c in cluster_centers]

        for n in ns:
            idx = closest(cluster_centers, n)
            clusters[idx].append(n)

        # FIXME HACK TODO: is it okay for there to be empty clusters?
        clusters = [c for c in clusters if len(c) > 0]

        for c in clusters:
            if (len(c) == 0):
                print('Errorrr....')
                print('Nodes: %d, centers: %s' % (
                    len(ns), repr(cluster_centers)))

            assert len(c) > 0

        new_cluster_centers = [center_of_gravity(c) for c in clusters]
        if new_cluster_centers == cluster_centers:
            root.stats['avg_kmeans_iter_f'] = float(
                root.stats['sum_kmeans_iter_f']
                / root.stats['count_kmeans_iter_f'])
            root.stats['longest_kmeans'] = max(
                root.stats['longest_kmeans'], (time.time() - t))
            return clusters
        else:
            cluster_centers = new_cluster_centers



import numpy as np

# define conversion matrices
_rgb2lms = np.array([[0.3811, 0.5783, 0.0402],
                     [0.1967, 0.7244, 0.0782],
                     [0.0241, 0.1288, 0.8444]])

_lms2lab = np.dot(
    np.array([[1 / (3**0.5), 0, 0],
              [0, 1 / (6**0.5), 0],
              [0, 0, 1 / (2**0.5)]]),
    np.array([[1, 1, 1],
              [1, 1, -2],
              [1, -1, 0]]),
)


def rgb_to_lab(im_rgb):
    """Transforms an image from RGB to LAB color space

    Parameters
    ----------
    im_rgb : array_like
        An RGB image

    Returns
    -------
    im_lab : array_like
        LAB representation of the input image `im_rgb`.

    See Also
    --------
    histomicstk.preprocessing.color_conversion.lab_to_rgb,
    histomicstk.preprocessing.color_normalization.reinhard

    References
    ----------
    .. [#] D. Ruderman, T. Cronin, and C. Chiao, "Statistics of cone
       responses to natural images: implications for visual coding,"
       J. Opt. Soc. Am. A vol.15, pp.2036-2045, 1998.

    """
    # get input image dimensions
    m = im_rgb.shape[0]
    n = im_rgb.shape[1]

    # calculate im_lms values from RGB
    im_rgb = np.reshape(im_rgb, (m * n, 3))
    im_lms = np.dot(_rgb2lms, np.transpose(im_rgb))
    im_lms[im_lms == 0] = np.spacing(1)

    # calculate LAB values from im_lms
    im_lab = np.dot(_lms2lab, np.log(im_lms))

    # reshape to 3-channel image
    im_lab = np.reshape(im_lab.transpose(), (m, n, 3))

    return im_lab



import numpy as np


def compact(im_label, compaction=3):
    """Performs a thinning operation on a label image to remove thin
    protrusions from objects that are in contact with the background. Applies a
    distance transform and sequentially removes pixels with small distances to
    background that are not connected to higher distance pixels.

    Parameters
    ----------
    im_label : array_like
        A labeled segmentation mask
    compaction : int
        Factor used in compacting objects to remove thin protrusions. Referred
        to as d in the reference below. Default value = 3.

    Notes
    -----
    Implemented from the reference below.

    Returns
    -------
    im_compact : array_like
        A labeled segmentation mask with thin protrusions removed.

    See Also
    --------
    histomicstk.segmentation.label.area_open,
    histomicstk.segmentation.label.condense,
    histomicstk.segmentation.label.shuffle,
    histomicstk.segmentation.label.split,
    histomicstk.segmentation.label.width_open

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
           Virtual Microscopy Images: A Minimum-Model Approach" in Nature
           Scientific Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    import scipy.ndimage as ndi

    # copy input image
    im_compact = im_label.copy()

    # generate distance map of label image
    D = ndi.distance_transform_cdt(im_compact > 0, metric='taxicab')

    # define 4-neighbors filtering kernel
    Kernel = np.zeros((3, 3), dtype=bool)
    Kernel[1, :] = True
    Kernel[:, 1] = True

    # sweep over distance values from d-1 to 1
    for i in np.arange(compaction - 1, 0, -1):

        # four-neighbor maxima of distance transform
        MaxD = ndi.maximum_filter(D, footprint=Kernel)

        # identify pixels whose max 4-neighbor is less than i+1
        Decrement = (D == i) & (MaxD < i + 1)

        # decrement non-compact pixels
        D[Decrement] -= 1

    # zero label pixels where D == 0
    im_compact[D == 0] = 0

    return im_compact



"""
# Modified from: https://code.google.com/archive/p/pyrtree/ .

# Copyright Google
# Under The 3-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION). HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""

import math


class Rect:
    """
    A rectangle class that stores: .

    an axis aligned rectangle, and: two flags (swapped_x and swapped_y).
    (The flags are stored implicitly via swaps in the order of
    minx/y and maxx/y.)
    """

    __slots__ = ('x', 'y', 'xx', 'yy', 'swapped_x', 'swapped_y')

    def __init__(self, minx, miny, maxx, maxy):
        """Placeholder."""
        self.swapped_x = (maxx < minx)
        self.swapped_y = (maxy < miny)
        self.x = minx
        self.y = miny
        self.xx = maxx
        self.yy = maxy

        if self.swapped_x:
            self.x, self.xx = maxx, minx
        if self.swapped_y:
            self.y, self.yy = maxy, miny

    def coords(self):
        """Placeholder."""
        return self.x, self.y, self.xx, self.yy

    def overlap(self, orect):
        """Placeholder."""
        return self.intersect(orect).area()

    def write_raw_coords(self, toarray, idx):
        """Placeholder."""
        toarray[idx] = self.x
        toarray[idx + 1] = self.y
        toarray[idx + 2] = self.xx
        toarray[idx + 3] = self.yy
        if (self.swapped_x):
            toarray[idx] = self.xx
            toarray[idx + 2] = self.x
        if (self.swapped_y):
            toarray[idx + 1] = self.yy
            toarray[idx + 3] = self.y

    def area(self):
        """Placeholder."""
        w = self.xx - self.x
        h = self.yy - self.y
        return w * h

    def extent(self):
        """Placeholder."""
        x = self.x
        y = self.y
        return (x, y, self.xx - x, self.yy - y)

    def grow(self, amt):
        """Placeholder."""
        a = amt * 0.5
        return Rect(self.x - a, self.y - a, self.xx + a, self.yy + a)

    def intersect(self, o):
        """Placeholder."""
        if self is NullRect:
            return NullRect
        if o is NullRect:
            return NullRect

        nx, ny = max(self.x, o.x), max(self.y, o.y)
        nx2, ny2 = min(self.xx, o.xx), min(self.yy, o.yy)
        w, h = nx2 - nx, ny2 - ny

        if w <= 0 or h <= 0:
            return NullRect

        return Rect(nx, ny, nx2, ny2)

    def does_contain(self, o):
        """Placeholder."""
        return self.does_containpoint((o.x, o.y)) and self.does_containpoint(
            (o.xx, o.yy))

    def does_intersect(self, o):
        """Placeholder."""
        return (self.intersect(o).area() > 0)

    def does_containpoint(self, p):
        """Placeholder."""
        x, y = p
        return (x >= self.x and x <= self.xx and y >= self.y and y <= self.yy)

    def union(self, o):
        """Placeholder."""
        if o is NullRect:
            return Rect(self.x, self.y, self.xx, self.yy)
        if self is NullRect:
            return Rect(o.x, o.y, o.xx, o.yy)

        x = self.x
        y = self.y
        xx = self.xx
        yy = self.yy
        ox = o.x
        oy = o.y
        oxx = o.xx
        oyy = o.yy

        nx = x if x < ox else ox
        ny = y if y < oy else oy
        nx2 = xx if xx > oxx else oxx
        ny2 = yy if yy > oyy else oyy

        res = Rect(nx, ny, nx2, ny2)

        return res

    def union_point(self, o):
        """Placeholder."""
        x, y = o
        return self.union(Rect(x, y, x, y))

    def diagonal_sq(self):
        """Placeholder."""
        if self is NullRect:
            return 0
        w = self.xx - self.x
        h = self.yy - self.y
        return w * w + h * h

    def diagonal(self):
        """Placeholder."""
        return math.sqrt(self.diagonal_sq())


NullRect = Rect(0.0, 0.0, 0.0, 0.0)
NullRect.swapped_x = False
NullRect.swapped_y = False


def union_all(kids):
    """Placeholder."""
    cur = NullRect
    for k in kids:
        cur = cur.union(k.rect)
    assert not cur.swapped_x
    return cur



"""Placeholder."""

import numpy as np


def rgb_to_hsi(im):
    """Convert to HSI the RGB pixels in im.

    Adapted from
    https://en.wikipedia.org/wiki/HSL_and_HSV#Hue_and_chroma.

    """
    im = np.moveaxis(im, -1, 0)
    if len(im) not in (3, 4):
        msg = (
            'Expected 3-channel RGB or 4-channel RGBA image;'
            ' received a {}-channel image'.format(len(im))
        )
        raise ValueError(msg)
    im = im[:3]
    hues = (np.arctan2(3**0.5 * (im[1] - im[2]),
                       2 * im[0] - im[1] - im[2]) / (2 * np.pi)) % 1
    intensities = im.mean(0)
    saturations = np.where(
        intensities, 1 - im.min(0) / np.maximum(intensities, 1e-10), 0)
    return np.stack([hues, saturations, intensities], -1)



from ..color_conversion import rgb_to_sda
from .separate_stains_xu_snmf import separate_stains_xu_snmf


def rgb_separate_stains_xu_snmf(im_rgb, I_0, *args, **kwargs):
    """Compute the stain matrix for color deconvolution with SNMF from an
    RGB image or matrix.

    Parameters
    ----------
    im_rgb : array_like
        Image (MxNx3) or matrix (3xN) in RGB space for which to compute the
        stain matrix.
    I_0 : float or array_like
        Per-channel background intensities, or one intensity to use for all
        channels if a float.

    Returns
    -------
    w : array_like
        A 3x3 matrix of stain column vectors, in SDA space

    Note
    ----
    This is a thin wrapper around separate_stains_xu_snmf and as such
    the w_init parameter must be manually passed.

    For additional input arguments and documentation, please see
    histomicstk.preprocessing.color_deconvolution.separate_stains_xu_snmf.
    im_sda is computed and passed as part of this routine.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.separate_stains_xu_snmf

    """
    im_sda = rgb_to_sda(im_rgb, I_0)
    return separate_stains_xu_snmf(im_sda, *args, **kwargs)



import numpy as np


def complement_stain_matrix(w):
    """Generates a complemented stain matrix
    Used to fill out empty columns of a stain matrix for use with
    color_deconvolution. Replaces right-most column with normalized
    cross-product of first two columns.

    Parameters
    ----------
    w : array_like
        A 3x3 stain calibration matrix with stain color vectors in columns.

    Returns
    -------
    w_comp : array_like
        A 3x3 complemented stain calibration matrix with a third
        orthogonal column.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.color_deconvolution

    """
    stain0 = w[:, 0]
    stain1 = w[:, 1]
    stain2 = np.cross(stain0, stain1)
    # Normalize new vector to have unit norm
    return np.array([stain0, stain1, stain2 / np.linalg.norm(stain2)]).T



import importlib
import sys

from setuptools import find_packages

try:
    from skbuild import setup
except ImportError:
    sys.stderr.write("""scikit-build is required to build from source or run tox.
Please run:
  python -m pip install scikit-build
""")
    # from setuptools import setup
    sys.exit(1)


with open('README.rst') as readme_file:
    readme = readme_file.read()


setup(
    name='histomicstk',
    use_scm_version={'local_scheme': 'no-local-version',
                     'fallback_version': '0.0.0'},
    description='A Python toolkit for Histopathology Image Analysis',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Kitware, Inc.',
    author_email='developers@digitalslidearchive.net',
    url='https://github.com/DigitalSlideArchive/HistomicsTK',
    packages=find_packages(exclude=['tests', '*_test*']),
    package_dir={
        'histomicstk': 'histomicstk',
    },
    include_package_data=True,
    install_requires=[
        'girder-client',
        # scientific packages
        'nimfa',
        'numpy',
        'scipy',
        'Pillow',
        'pandas',
        'scikit-image',
        'scikit-learn',
        'imageio',
        'shapely',
        'sqlalchemy',
        'matplotlib',
        'pyvips',
        # dask packages
        'dask[dataframe]',
        'distributed',
        # large image; for non-linux systems only install the PIL tile source
        # by default.
        'large-image[sources];sys.platform=="linux" or sys.platform=="linux2"',
        'large-image[common];sys.platform!="linux" and sys.platform!="linux2"',
        'large-image-converter;sys.platform=="linux" or sys.platform=="linux2"',
        'girder-slicer-cli-web',
        # cli
        'ctk-cli',
    ] + (
        # Only require opencv if it isn't installed.  This can allow alternate
        # forms to be in the environment (such as a headed form) without
        # causing conflicts
        [
            'opencv-python-headless',
        ] if not importlib.util.find_spec('cv2') else []
    ),
    license='Apache Software License 2.0',
    keywords='histomicstk',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': ['histomicstk = histomicstk.cli.__main__:main'],
    },
    python_requires='>=3.8',
)



import collections

import numpy as np

import histomicstk as htk


def gaussian_voting(I, rmax=35, rmin=10, sSigma=5, Tau=5, bw=15, Psi=0.3):
    """Performs nuclear detection using Gaussian kernel voting.

    Uses a gaussian kernel to localize the centroids of cell nuclei. Takes
    as input a hematoxylin-deconvolved image and uses the gradient signal to
    cast directed votes towards the center of cell nuclei. These votes are
    blurred by a gaussian kernel, and are spatially clustered using the
    mean-shift algorithm. Convolutions are performed separably to reduce
    compute time.

    Parameters
    ----------
    I : array_like
        A hematoxylin intensity image obtained from ColorDeconvolution.
    rmax : double
        Upper-limit on voting area extent. Default value = 35.
    rmin : double
        Lower-limit on voting area extent. Default value = 10.
    sSigma : double
        Standard deviation of smoothing kernel used in gradient calculation.
        Default value = 5.
    Tau : double
        Lower limit on gradient magnitude for casting a vote. Default
        value = 5.
    bw : double
        Bandwidth parameter for mean-shift clustering. Default value = 15.
    Psi : double
        Lower limit threshold on votes. Expressed as a fraction of
        the maximum vote, ranges from [0, 1). Default value = 0.3.

    Returns
    -------
    Centroids : array_like
        An N x 2 array defining the (x,y) coordinates of cell centroids.
    Votes : array_like
        An intensity image containing the blurred votes obtained by voting.

    Notes
    -----
    Return values are returned as a namedtuple

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.color_deconvolution

    References
    ----------
    .. [#] X. Qi, F. Xing, D.J. Foran, L. Yang, "Robust Segmentation of
       Overlapping Cells in Histopathology Specimens Using Parallel
       Seed Detection and Repulsive Level Set," in IEEE Transactions
       on Biomedical Engineering, vol.59,no.23,pp.754-65, 2011.

    """
    import scipy.signal as signal
    import sklearn.cluster as cl

    # calculate standard deviation of voting kernel
    vSigma = (rmax - rmin) / 3

    # calculate voting radius
    r = int(round((rmax + rmin) / 2))

    # generate separable gaussian derivative kernels
    Grad = htk.filters.edge.gaussian_grad(I, sSigma)
    dMag = (Grad.dX**2 + Grad.dY**2)**0.5

    # threshold gradient image to identify voting pixels
    dMask = dMag >= Tau
    Voting = dMask.nonzero()

    # initialize voting field - pad by 'r' along each edge
    Votes = np.zeros((I.shape[0] + 2 * r, I.shape[1] + 2 * r))
    for i in range(Voting[0].size):

        # calculate center point of voting region
        mux = round(Voting[1][i] + r * Grad.dX[Voting[0][i]][Voting[1][i]] /
                    dMag[Voting[0][i]][Voting[1][i]])
        muy = round(Voting[0][i] + r * Grad.dY[Voting[0][i]][Voting[1][i]] /
                    dMag[Voting[0][i]][Voting[1][i]])

        # enter weighted votes at these locations
        Votes[r + muy, r + mux] = Votes[r + muy, r + mux] + \
            dMag[Voting[0][i]][Voting[1][i]]

    # create voting kernel
    x = np.linspace(0, np.ceil(2 * 3 * vSigma), int(np.ceil(2 * 3 * vSigma) + 1))
    y = np.linspace(0, np.ceil(2 * 3 * vSigma), int(np.ceil(2 * 3 * vSigma) + 1))
    x -= np.ceil(2 * 3 * vSigma) / 2  # center at zero
    y -= np.ceil(2 * 3 * vSigma) / 2
    x = np.reshape(x, (1, x.size))  # reshape to 2D row and column vectors
    y = np.reshape(y, (y.size, 1))
    xK = np.exp(-x**2 / (2 * vSigma**2)) / (sSigma * (2 * np.pi) ** 0.5)
    yK = np.exp(-y**2 / (2 * vSigma**2)) / (sSigma * (2 * np.pi) ** 0.5)

    # perform convolutions with voting kernel for vote-smoothing
    Votes = signal.convolve2d(Votes, xK, mode='full', boundary='fill')
    Votes = signal.convolve2d(Votes, yK, mode='full', boundary='fill')

    # crop voting image to size of original input image
    Votes = Votes[r + int(np.floor(x.size / 2)):-(r + int(np.ceil(x.size / 2))),
                  r + int(np.floor(x.size / 2)):-(r + int(np.ceil(x.size / 2)))]

    # generate sets of potential seed points
    Seeds = [None] * np.arange(np.floor(10 * Psi) / 10, 0.9, 0.1).size
    for i, p in enumerate(np.arange(np.floor(10 * Psi) / 10, 0.9, 0.1)):
        Points = np.nonzero(Votes >= p * Votes.max())
        Seeds[i] = np.column_stack((Points[0].transpose(),
                                    Points[1].transpose()))

    # concatenate seed point lists
    Seeds = np.vstack(Seeds)

    # run mean-shift algorithm to collect
    ms = cl.MeanShift(bandwidth=bw, bin_seeding=True)
    ms.fit(Seeds)

    # build output tuple
    Output = collections.namedtuple('Output', ['X', 'Y'])
    Nuclei = Output(ms.cluster_centers_[:, 1], ms.cluster_centers_[:, 0])

    return Nuclei, Votes



import numpy as np

from ._max_clustering_cython import _max_clustering_cython


def max_clustering(im_response, im_fgnd_mask, r=10):
    """Local max clustering pixel aggregation for nuclear segmentation.
    Takes as input a constrained log or other filtered nuclear image, a binary
    nuclear mask, and a clustering radius. For each pixel in the nuclear mask,
    the local max is identified. A hierarchy of local maxima is defined, and
    the root nodes used to define the label image.

    Parameters
    ----------
    im_response : array_like
        A filtered-smoothed image where the maxima correspond to nuclear
        center. Typically obtained by constrained-LoG filtering on a
        hematoxylin intensity image obtained from ColorDeconvolution.
    im_fgnd_mask : array_like
        A binary mask of type boolean where nuclei pixels have value
        'True', and non-nuclear pixels have value 'False'.
    r : float
        A scalar defining the clustering radius. Default value = 10.

    Returns
    -------
    im_label : array_like
        im_label image where positive values correspond to foreground pixels that
        share mutual sinks.
    seeds : array_like
        An N x 2 array defining the (x,y) coordinates of nuclei seeds.
    max_response : array_like
        An N x 1 array containing the maximum response value corresponding to
        'seeds'.

    See Also
    --------
    histomicstk.filters.shape.clog

    References
    ----------
    .. [#] XW. Wu et al "The local maximum clustering method and its
       application in microarray gene expression data analysis,"
       EURASIP J. Appl. Signal Processing, volume 2004, no.1, pp.53-63,
       2004.
    .. [#] Y. Al-Kofahi et al "Improved Automatic Detection and Segmentation
       of Cell Nuclei in Histopathology Images" in IEEE Transactions on
       Biomedical Engineering,vol.57,no.4,pp.847-52, 2010.

    """
    import skimage.measure

    # find local maxima of all foreground pixels
    mval, mind = _max_clustering_cython(
        im_response, im_fgnd_mask.astype(np.int32), r,
    )

    # identify connected regions of local maxima and define their seeds
    im_label = skimage.measure.label(im_fgnd_mask & (im_response == mval))

    if not np.any(im_label):
        return im_label, None, None

    # compute normalized response
    min_resp = im_response.min()
    max_resp = im_response.max()
    resp_range = max_resp - min_resp

    if resp_range == 0:
        return np.zeros_like(im_label), None, None

    im_response_nmzd = (im_response - min_resp) / resp_range

    # compute object properties
    obj_props = skimage.measure.regionprops(im_label, im_response_nmzd)

    obj_props = [prop for prop in obj_props if np.isfinite(prop.weighted_centroid).all()]

    num_labels = len(obj_props)

    if num_labels == 0:
        return im_label, None, None

    # extract object seeds
    seeds = np.array(
        [obj_props[i].weighted_centroid for i in range(num_labels)])
    seeds = np.round(seeds).astype(int)

    # fix seeds outside the object region - happens for non-convex objects
    for i in range(num_labels):

        sy = seeds[i, 0]
        sx = seeds[i, 1]

        if im_label[sy, sx] == obj_props[i].label:
            continue

        # find object point with closest manhattan distance to center of mass
        pts = obj_props[i].coords

        ydist = np.abs(pts[:, 0] - sy)
        xdist = np.abs(pts[:, 1] - sx)

        seeds[i, :] = pts[np.argmin(xdist + ydist), :]

        assert im_label[seeds[i, 0], seeds[i, 1]] == obj_props[i].label

    # get seed responses
    max_response = im_response[seeds[:, 0], seeds[:, 1]]

    # set label of each foreground pixel to the label of its nearest peak
    im_label_flat = im_label.ravel()

    pind = np.flatnonzero(im_fgnd_mask)

    mind_flat = mind.ravel()

    im_label_flat[pind] = im_label_flat[mind_flat[pind]]

    # return
    return im_label, seeds, max_response



import numpy

from histomicstk.utils import sample_pixels


def background_intensity(*args, **kwargs):
    """Sample the background of the slide identified by slide_path to
    compute the background intensities in the RGB channels.  Arguments
    are as in histomicstk.utils.sample_pixels, with background forced
    to True.

    Notes
    -----
    The `magnification` parameter defaults to 1.25x, instead of the
    native scan magnification as in sample_pixels.

    """
    if 'background' in kwargs:
        msg = '"background" argument may not be specified'
        raise ValueError(msg)

    kwargs['background'] = True

    if 'magnification' not in kwargs:
        kwargs['magnification'] = 1.25

    sample = sample_pixels(*args, **kwargs)

    return numpy.median(sample, axis=0)
